# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from regions import (
    CircleSkyRegion,
    EllipseAnnulusSkyRegion,
    EllipseSkyRegion,
    PointSkyRegion,
    RectangleSkyRegion,
)
from gammapy.data import DataStore
from gammapy.datasets import SpectrumDataset
from gammapy.makers import (
    ReflectedRegionsBackgroundMaker,
    ReflectedRegionsFinder,
    SafeMaskMaker,
    SpectrumDatasetMaker,
    WobbleRegionsFinder,
)
from gammapy.maps import Map, MapAxis, RegionGeom, WcsGeom
from gammapy.utils.regions import compound_region_to_regions
from gammapy.utils.testing import assert_quantity_allclose, requires_data


@pytest.fixture(scope="session")
def exclusion_mask():
    """Example mask for testing."""
    pos = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    exclusion_region = CircleSkyRegion(pos, Angle(0.3, "deg"))
    geom = WcsGeom.create(skydir=pos, binsz=0.02, width=10.0)
    return ~geom.region_mask([exclusion_region])


@pytest.fixture(scope="session")
def on_region():
    """Example on_region for testing."""
    pos = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    radius = Angle(0.11, "deg")
    region = CircleSkyRegion(pos, radius)
    return region


@pytest.fixture(scope="session")
def observations():
    """Example observation list for testing."""
    datastore = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    obs_ids = [23523, 23526]
    return datastore.get_observations(obs_ids)


@pytest.fixture(scope="session")
def observations_fixed_rad_max():
    """Example observation list for testing."""
    datastore = DataStore.from_dir("$GAMMAPY_DATA/joint-crab/dl3/magic/")
    obs_ids = [5029748]
    return datastore.get_observations(obs_ids, required_irf="point-like")


@pytest.fixture()
def reflected_bkg_maker(exclusion_mask):
    finder = ReflectedRegionsFinder()
    return ReflectedRegionsBackgroundMaker(
        region_finder=finder,
        exclusion_mask=exclusion_mask,
    )


region_finder_param = [
    (SkyCoord(83.2, 22.5, unit="deg"), 15, Angle("82.592 deg"), 17, 17),
    (SkyCoord(84.2, 22.5, unit="deg"), 17, Angle("83.636 deg"), 19, 19),
    (SkyCoord(83.2, 21.5, unit="deg"), 15, Angle("83.672 deg"), 17, 17),
]


@requires_data()
@pytest.mark.parametrize(
    "pointing_pos, nreg1, reg3_ra, nreg2, nreg3", region_finder_param
)
def test_find_reflected_regions(
    exclusion_mask, on_region, pointing_pos, nreg1, reg3_ra, nreg2, nreg3
):
    pointing = pointing_pos
    finder = ReflectedRegionsFinder(
        min_distance_input="0 deg",
    )
    regions, _ = finder.run(
        center=pointing,
        region=on_region,
        exclusion_mask=exclusion_mask,
    )
    assert len(regions) == nreg1
    assert_quantity_allclose(regions[3].center.icrs.ra, reg3_ra, rtol=1e-2)

    # Test without exclusion
    regions, _ = finder.run(center=pointing, region=on_region)
    assert len(regions) == nreg2

    # Test with too small exclusion
    small_mask = exclusion_mask.cutout(pointing, Angle("0.1 deg"))
    regions, _ = finder.run(
        center=pointing,
        region=on_region,
        exclusion_mask=small_mask,
    )
    assert len(regions) == nreg3

    # Test with maximum number of regions
    finder.max_region_number = 5
    regions, _ = finder.run(
        center=pointing,
        region=on_region,
        exclusion_mask=small_mask,
    )
    assert len(regions) == 5

    # Test with an other type of region
    on_ellipse_annulus = EllipseAnnulusSkyRegion(
        center=on_region.center.galactic,
        inner_width=0.1 * u.deg,
        outer_width=0.2 * u.deg,
        inner_height=0.3 * u.deg,
        outer_height=0.6 * u.deg,
        angle=130 * u.deg,
    )
    regions, _ = finder.run(
        region=on_ellipse_annulus,
        center=pointing,
        exclusion_mask=small_mask,
    )
    assert len(regions) == 5


center = SkyCoord(0.5, 0.0, unit="deg")
other_region_finder_param = [
    (RectangleSkyRegion(center, 0.5 * u.deg, 0.5 * u.deg, angle=0 * u.deg), 3),
    (RectangleSkyRegion(center, 0.5 * u.deg, 1 * u.deg, angle=0 * u.deg), 1),
    (RectangleSkyRegion(center, 0.5 * u.deg, 1 * u.deg, angle=90 * u.deg), 1),
    (EllipseSkyRegion(center, 0.1 * u.deg, 1 * u.deg, angle=0 * u.deg), 2),
    (EllipseSkyRegion(center, 0.1 * u.deg, 1 * u.deg, angle=60 * u.deg), 3),
    (EllipseSkyRegion(center, 0.1 * u.deg, 1 * u.deg, angle=90 * u.deg), 2),
]


@pytest.mark.parametrize("region, nreg", other_region_finder_param)
def test_non_circular_regions(region, nreg):
    pointing = SkyCoord(0.0, 0.0, unit="deg")

    finder = ReflectedRegionsFinder(min_distance_input="0 deg")
    regions, _ = finder.run(center=pointing, region=region)
    assert len(regions) == nreg


def test_bad_on_region(exclusion_mask, on_region):
    pointing = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    finder = ReflectedRegionsFinder(
        min_distance_input="0 deg",
    )
    regions, _ = finder.run(
        center=pointing,
        region=on_region,
        exclusion_mask=exclusion_mask,
    )
    assert len(regions) == 0


@requires_data()
def test_reflected_bkg_maker(on_region, reflected_bkg_maker, observations):
    datasets = []

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")
    e_true = MapAxis.from_edges(np.logspace(-0.5, 2, 11) * u.TeV, name="energy_true")

    geom = RegionGeom(region=on_region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    maker = SpectrumDatasetMaker(selection=["counts"])

    for obs in observations:
        dataset = maker.run(dataset_empty, obs)
        dataset_on_off = reflected_bkg_maker.run(dataset, obs)
        datasets.append(dataset_on_off)

    assert_allclose(datasets[0].counts_off.data.sum(), 76)
    assert_allclose(datasets[1].counts_off.data.sum(), 60)

    regions_0 = compound_region_to_regions(datasets[0].counts_off.geom.region)
    regions_1 = compound_region_to_regions(datasets[1].counts_off.geom.region)
    assert_allclose(len(regions_0), 11)
    assert_allclose(len(regions_1), 11)


@requires_data()
def test_reflected_bkg_maker_no_off(reflected_bkg_maker, observations, caplog):
    pos = SkyCoord(83.6333313, 21.51444435, unit="deg", frame="icrs")
    radius = Angle(0.11, "deg")
    region = CircleSkyRegion(pos, radius)

    maker = SpectrumDatasetMaker(selection=["counts", "exposure"])

    datasets = []

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")
    e_true = MapAxis.from_edges(np.logspace(-0.5, 2, 11) * u.TeV, name="energy_true")
    geom = RegionGeom.create(region=region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    safe_mask_masker = SafeMaskMaker(methods=["aeff-max"], aeff_percent=10)
    for obs in observations:
        dataset = maker.run(dataset_empty, obs)
        dataset_on_off = reflected_bkg_maker.run(dataset, obs)
        dataset_on_off = safe_mask_masker.run(dataset_on_off, obs)
        datasets.append(dataset_on_off)

    assert datasets[0].counts_off is None
    assert_allclose(datasets[0].acceptance_off, 0)
    assert_allclose(datasets[0].mask_safe.data, False)

    assert "WARNING" in [record.levelname for record in caplog.records]

    message1 = (
        f"ReflectedRegionsBackgroundMaker failed. "
        f"No OFF region found outside exclusion mask for dataset '{datasets[0].name}'."
    )
    message2 = (
        f"ReflectedRegionsBackgroundMaker failed. "
        f"Setting {datasets[0].name} mask to False."
    )

    assert message1 in [record.message for record in caplog.records]
    assert message2 in [record.message for record in caplog.records]


@requires_data()
def test_reflected_bkg_maker_no_off_background(reflected_bkg_maker, observations):
    pos = SkyCoord(83.6333313, 21.51444435, unit="deg", frame="icrs")
    radius = Angle(0.11, "deg")
    region = CircleSkyRegion(pos, radius)

    maker = SpectrumDatasetMaker(selection=["counts", "background"])

    datasets = []

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")
    e_true = MapAxis.from_edges(np.logspace(-0.5, 2, 11) * u.TeV, name="energy_true")
    geom = RegionGeom.create(region=region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    for obs in observations:
        dataset = maker.run(dataset_empty, obs)
        dataset_on_off = reflected_bkg_maker.run(dataset, obs)
        datasets.append(dataset_on_off)

    assert_allclose(datasets[0].counts_off.data, 0)
    assert_allclose(datasets[0].acceptance_off, 0)


def test_wobble_regions_finder():
    center = SkyCoord(83.6333313, 21.51444435, unit="deg", frame="icrs")
    source_angle = 35 * u.deg

    on_region = CircleSkyRegion(
        center=center,
        radius=0.15 * u.deg,
    )
    pointing = on_region.center.directional_offset_by(
        separation=0.4 * u.deg,
        position_angle=180 * u.deg + source_angle,
    )

    assert u.isclose(
        pointing.position_angle(on_region.center).to(u.deg), source_angle, rtol=0.05
    )
    n_off_regions = 3

    finder = WobbleRegionsFinder(n_off_regions)
    regions, _ = finder.run(on_region, pointing)

    assert len(regions) == 3

    for i, off_region in enumerate(regions, start=1):
        assert u.isclose(pointing.separation(off_region.center), 0.4 * u.deg)
        expected = source_angle + 360 * u.deg * i / (n_off_regions + 1)
        assert u.isclose(
            pointing.position_angle(off_region.center).to(u.deg),
            expected.to(u.deg),
            rtol=0.001,
        )

    # test with exclusion region
    pos = pointing.directional_offset_by(
        separation=0.4 * u.deg,
        position_angle=source_angle + 90 * u.deg,
    )
    exclusion_region = CircleSkyRegion(pos, Angle(0.2, "deg"))
    geom = WcsGeom.create(skydir=pos, binsz=0.01, width=10.0)
    exclusion_mask = ~geom.region_mask([exclusion_region])

    finder = WobbleRegionsFinder(n_off_regions)
    regions, _ = finder.run(on_region, pointing, exclusion_mask)

    assert len(regions) == 2


def test_wobble_regions_finder_overlapping(caplog):
    """Test that overlapping regions are not produced"""
    center = SkyCoord(83.6333313, 21.51444435, unit="deg", frame="icrs")
    source_angle = 35 * u.deg

    on_region = CircleSkyRegion(
        center=center,
        radius=0.5 * u.deg,
    )
    pointing = on_region.center.directional_offset_by(
        separation=0.4 * u.deg,
        position_angle=180 * u.deg + source_angle,
    )

    n_off_regions = 3

    finder = WobbleRegionsFinder(n_off_regions)

    with caplog.at_level(logging.WARNING):
        regions, _ = finder.run(on_region, pointing)

    assert len(regions) == 0
    assert caplog.record_tuples == [
        (
            "gammapy.makers.background.reflected",
            logging.WARNING,
            "Found overlapping off regions, returning no regions",
        )
    ]


@requires_data()
def test_reflected_bkg_maker_with_wobble_finder(
    on_region, observations, exclusion_mask
):
    datasets = []

    reflected_bkg_maker = ReflectedRegionsBackgroundMaker(
        region_finder=WobbleRegionsFinder(n_off_regions=3),
        exclusion_mask=exclusion_mask,
    )

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")

    geom = RegionGeom(region=on_region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom)

    spectrum_dataset_maker = SpectrumDatasetMaker(selection=["counts"])

    for obs in observations:
        dataset = spectrum_dataset_maker.run(dataset_empty, obs)
        dataset_on_off = reflected_bkg_maker.run(dataset, obs)
        datasets.append(dataset_on_off)

    regions_0 = compound_region_to_regions(datasets[0].counts_off.geom.region)
    regions_1 = compound_region_to_regions(datasets[1].counts_off.geom.region)
    assert_allclose(len(regions_0), 3)
    assert_allclose(len(regions_1), 3)


@requires_data()
def test_reflected_bkg_maker_fixed_rad_max(
    reflected_bkg_maker, observations_fixed_rad_max
):
    e_reco = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV")
    e_true = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV", name="energy_true")

    pos = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    radius = Angle(0.1414, "deg")
    region = CircleSkyRegion(pos, radius)

    geom = RegionGeom(region=region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    maker = SpectrumDatasetMaker(selection=["counts"])

    obs = observations_fixed_rad_max[0]
    dataset = maker.run(dataset_empty, obs)
    dataset_on_off = reflected_bkg_maker.run(dataset, obs)

    assert_allclose(dataset_on_off.counts_off.data.sum(), 217)

    regions_0 = compound_region_to_regions(dataset_on_off.counts_off.geom.region)
    assert_allclose(len(regions_0), 6)


@requires_data()
def test_reflected_bkg_maker_fixed_rad_max_wobble(
    exclusion_mask, observations_fixed_rad_max
):
    reflected_bkg_maker = ReflectedRegionsBackgroundMaker(
        region_finder=WobbleRegionsFinder(n_off_regions=3),
        exclusion_mask=exclusion_mask,
    )
    e_reco = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV")
    e_true = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV", name="energy_true")

    pos = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    radius = Angle(0.1414, "deg")
    region = CircleSkyRegion(pos, radius)

    geom = RegionGeom(region=region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    maker = SpectrumDatasetMaker(selection=["counts"])

    obs = observations_fixed_rad_max[0]
    dataset = maker.run(dataset_empty, obs)
    dataset_on_off = reflected_bkg_maker.run(dataset, obs)

    assert_allclose(dataset_on_off.counts_off.data.sum(), 102)

    regions_0 = compound_region_to_regions(dataset_on_off.counts_off.geom.region)
    assert_allclose(len(regions_0), 3)


@requires_data()
def test_reflected_bkg_maker_fixed_rad_max_bad(
    reflected_bkg_maker, observations_fixed_rad_max
):
    e_reco = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV")

    pos = SkyCoord(83.63, 22.01, unit="deg", frame="icrs")
    bad_radius = Angle(0.11, "deg")
    region_bad_size = CircleSkyRegion(pos, bad_radius)

    maker = SpectrumDatasetMaker(selection=["counts"])
    obs = observations_fixed_rad_max[0]

    geom_bad_size = RegionGeom(region=region_bad_size, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom_bad_size)

    dataset = maker.run(dataset_empty, obs)
    with pytest.raises(ValueError):
        reflected_bkg_maker.run(dataset, obs)

    region_bad_shape = RectangleSkyRegion(pos, 0.2 * u.deg, 0.2 * u.deg)
    geom_bad_shape = RegionGeom(region_bad_shape, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom_bad_shape)

    dataset = maker.run(dataset_empty, obs)
    with pytest.raises(ValueError):
        reflected_bkg_maker.run(dataset, obs)

    region_point_shape = PointSkyRegion(pos)
    geom_point_shape = RegionGeom(region_point_shape, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom_point_shape)

    dataset = maker.run(dataset_empty, obs)
    with pytest.raises(TypeError):
        reflected_bkg_maker.run(dataset, obs)


def test_reflected_bkg_exclsion_error(exclusion_mask):

    energy = MapAxis.from_energy_bounds(1 * u.TeV, 2 * u.TeV, nbin=1, name="energy")

    exclusion_cube = exclusion_mask.to_cube([energy])
    ReflectedRegionsBackgroundMaker(exclusion_mask=exclusion_cube)

    energy = MapAxis.from_energy_bounds(2 * u.TeV, 3 * u.TeV, nbin=1, name="energy")
    exclusion_cube_2 = exclusion_mask.to_cube([energy])

    exclusion_stack = Map.from_stack([exclusion_cube, exclusion_cube_2])

    with pytest.raises(ValueError):
        ReflectedRegionsBackgroundMaker(exclusion_mask=exclusion_stack)

    exclusion_mask.data = exclusion_mask.data.astype("int")

    with pytest.raises(ValueError):
        ReflectedRegionsBackgroundMaker(exclusion_mask=exclusion_mask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import abc
import copy
import html
import inspect
import json
from collections import OrderedDict
from itertools import repeat
import numpy as np
from astropy import units as u
from astropy.io import fits
import matplotlib.pyplot as plt
import gammapy.utils.parallel as parallel
from gammapy.utils.compat import COPY_IF_NEEDED
from gammapy.utils.random import InverseCDFSampler, get_random_state
from gammapy.utils.scripts import make_path
from gammapy.utils.types import JsonQuantityDecoder
from gammapy.utils.units import energy_unit_format
from .axes import MapAxis
from .coord import MapCoord
from .geom import pix_tuple_to_idx

__all__ = ["Map"]

u.add_enabled_units([u.def_unit("transit", u.sday)])


class Map(abc.ABC):
    """Abstract map class.

    This can represent WCS or HEALPix-based maps
    with 2 spatial dimensions and N non-spatial dimensions.

    Parameters
    ----------
    geom : `~gammapy.maps.Geom`
        Geometry.
    data : `~numpy.ndarray` or `~astropy.units.Quantity`
        Data array.
    meta : `dict`, optional
        Dictionary to store metadata. Default is None.
    unit : str or `~astropy.units.Unit`, optional
        Data unit, ignored if data is a Quantity. Default is ''.
    """

    tag = "map"

    def __init__(self, geom, data, meta=None, unit=""):
        self._geom = geom

        if isinstance(data, u.Quantity):
            self._unit = u.Unit(unit)
            self.quantity = data
        else:
            self.data = data
            self._unit = u.Unit(unit)

        if meta is None:
            self.meta = {}
        else:
            self.meta = meta

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def _init_copy(self, **kwargs):
        """Init map instance by copying missing init arguments from self."""
        argnames = inspect.getfullargspec(self.__init__).args
        argnames.remove("self")
        argnames.remove("dtype")

        for arg in argnames:
            value = getattr(self, "_" + arg)
            if arg not in kwargs:
                kwargs[arg] = copy.deepcopy(value)

        return self.from_geom(**kwargs)

    @property
    def is_mask(self):
        """Whether map is a mask with boolean data type."""
        return self.data.dtype == bool

    @property
    def geom(self):
        """Map geometry as a `~gammapy.maps.Geom` object."""
        return self._geom

    @property
    def data(self):
        """Data array as a `~numpy.ndarray` object."""
        return self._data

    @data.setter
    def data(self, value):
        """Set data.

        Parameters
        ----------
        value : array-like
            Data array.
        """
        if np.isscalar(value):
            value = value * np.ones(self.geom.data_shape, dtype=type(value))

        if isinstance(value, u.Quantity):
            raise TypeError("Map data must be a Numpy array. Set unit separately")

        if not value.shape == self.geom.data_shape:
            try:
                value = np.broadcast_to(value, self.geom.data_shape, subok=True)
            except ValueError as exc:
                raise ValueError(
                    f"Input shape {value.shape} is not compatible with shape from geometry {self.geom.data_shape}"
                ) from exc

        self._data = value

    @property
    def unit(self):
        """Map unit as an `~astropy.units.Unit` object."""
        return self._unit

    @property
    def meta(self):
        """Map metadata as a `dict`."""
        return self._meta

    @meta.setter
    def meta(self, val):
        self._meta = val

    @property
    def quantity(self):
        """Map data as a `~astropy.units.Quantity` object."""
        return u.Quantity(self.data, self.unit, copy=COPY_IF_NEEDED)

    @quantity.setter
    def quantity(self, val):
        """Set data and unit.

        Parameters
        ----------
        val : `~astropy.units.Quantity`
           Quantity.
        """
        val = u.Quantity(val, copy=COPY_IF_NEEDED)

        self.data = val.value
        self._unit = val.unit

    def rename_axes(self, names, new_names):
        """Rename the Map axes.

        Parameters
        ----------
        names : str or list of str
            Names of the axes.
        new_names : str or list of str
            New names of the axes. The list must be of the same length as `names`).

        Returns
        -------
        geom : `~Map`
            Map with renamed axes.
        """
        geom = self.geom.rename_axes(names=names, new_names=new_names)
        return self._init_copy(geom=geom)

    @staticmethod
    def create(**kwargs):
        """Create an empty map object.

        This method accepts generic options listed below, as well as options
        for `HpxMap` and `WcsMap` objects. For WCS-specific options, see
        `WcsMap.create`; for HPX-specific options, see `HpxMap.create`.

        Parameters
        ----------
        frame : {"icrs", "galactic"}
            Coordinate system, either Galactic ("galactic") or Equatorial ("icrs").
        map_type : {'wcs', 'wcs-sparse', 'hpx', 'hpx-sparse', 'region'}
            Map type.  Selects the class that will be used to
            instantiate the map. Default is 'wcs'.
        binsz : float or `~numpy.ndarray`
            Pixel size in degrees.
        skydir : `~astropy.coordinates.SkyCoord`
            Coordinate of the map center.
        axes : list
            List of `~MapAxis` objects for each non-spatial dimension.
            If None then the map will be a 2D image.
        dtype : str
            Data type, default is 'float32'
        unit : str or `~astropy.units.Unit`
            Data unit.
        meta : `dict`
            Dictionary to store metadata.
        region : `~regions.SkyRegion`
            Sky region used for the region map.

        Returns
        -------
        map : `Map`
            Empty map object.
        """
        from .hpx import HpxMap
        from .region import RegionNDMap
        from .wcs import WcsMap

        map_type = kwargs.setdefault("map_type", "wcs")
        if "wcs" in map_type.lower():
            return WcsMap.create(**kwargs)
        elif "hpx" in map_type.lower():
            return HpxMap.create(**kwargs)
        elif map_type == "region":
            _ = kwargs.pop("map_type")
            return RegionNDMap.create(**kwargs)
        else:
            raise ValueError(f"Unrecognized map type: {map_type!r}")

    @staticmethod
    def read(
        filename,
        hdu=None,
        hdu_bands=None,
        map_type="auto",
        format=None,
        colname=None,
        checksum=False,
    ):
        """Read a map from a FITS file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            Name of the FITS file.
        hdu : str, optional
            Name or index of the HDU with the map data. Default is None.
        hdu_bands : str, optional
            Name or index of the HDU with the BANDS table.  If not
            defined this will be inferred from the FITS header of the
            map HDU. Default is None.
        map_type : {'auto', 'wcs', 'wcs-sparse', 'hpx', 'hpx-sparse', 'region'}
            Map type. Selects the class that will be used to
            instantiate the map. The map type should be consistent
            with the format of the input file. If map_type is 'auto'
            then an appropriate map type will be inferred from the
            input file. Default is 'auto'.
        colname : str, optional
            data column name to be used for HEALPix map.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        map_out : `Map`
            Map object.
        """
        with fits.open(make_path(filename), memmap=False, checksum=checksum) as hdulist:
            return Map.from_hdulist(
                hdulist, hdu, hdu_bands, map_type, format=format, colname=colname
            )

    @staticmethod
    def from_geom(geom, meta=None, data=None, unit="", dtype="float32"):
        """Generate an empty map from a `Geom` instance.

        Parameters
        ----------
        geom : `Geom`
            Map geometry.
        meta : `dict`, optional
            Dictionary to store metadata. Default is None.
        data : `numpy.ndarray`, optional
            Data array. Default is None.
        unit : str or `~astropy.units.Unit`
            Data unit.
        dtype : str, optional
            Data type. Default is 'float32'.

        Returns
        -------
        map_out : `Map`
            Map object.

        """
        from .hpx import HpxGeom
        from .region import RegionGeom
        from .wcs import WcsGeom

        if isinstance(geom, HpxGeom):
            map_type = "hpx"
        elif isinstance(geom, WcsGeom):
            map_type = "wcs"
        elif isinstance(geom, RegionGeom):
            map_type = "region"
        else:
            raise ValueError("Unrecognized geom type.")

        cls_out = Map._get_map_cls(map_type)
        return cls_out(geom, data=data, meta=meta, unit=unit, dtype=dtype)

    @staticmethod
    def from_hdulist(
        hdulist, hdu=None, hdu_bands=None, map_type="auto", format=None, colname=None
    ):
        """Create from a `astropy.io.fits.HDUList` object.

        Parameters
        ----------
        hdulist :  `~astropy.io.fits.HDUList`
            HDU list containing HDUs for map data and bands.
        hdu : str, optional
            Name or index of the HDU with the map data. Default is None.
        hdu_bands : str, optional
            Name or index of the HDU with the BANDS table. Default is None.
        map_type : {"auto", "wcs", "hpx", "region"}
            Map type. Default is "auto".
        format : {'gadf', 'fgst-ccube', 'fgst-template'}
            FITS format convention. Default is None.
        colname : str, optional
            Data column name to be used for HEALPix map. Default is None.

        Returns
        -------
        map_out : `Map`
            Map object.
        """
        if map_type == "auto":
            map_type = Map._get_map_type(hdulist, hdu)
        cls_out = Map._get_map_cls(map_type)
        if map_type == "hpx":
            return cls_out.from_hdulist(
                hdulist, hdu=hdu, hdu_bands=hdu_bands, format=format, colname=colname
            )
        else:
            return cls_out.from_hdulist(
                hdulist, hdu=hdu, hdu_bands=hdu_bands, format=format
            )

    @staticmethod
    def _get_meta_from_header(header):
        """Load metadata from a FITS header."""
        if "META" in header:
            return json.loads(header["META"], cls=JsonQuantityDecoder)
        else:
            return {}

    @staticmethod
    def _get_map_type(hdu_list, hdu_name):
        """Infer map type from a FITS HDU.

        Only read header, never data, to have good performance.
        """
        if hdu_name is None:
            # Find the header of the first non-empty HDU
            header = hdu_list[0].header
            if header["NAXIS"] == 0:
                header = hdu_list[1].header
        else:
            header = hdu_list[hdu_name].header

        if ("PIXTYPE" in header) and (header["PIXTYPE"] == "HEALPIX"):
            return "hpx"
        elif "CTYPE1" in header:
            return "wcs"
        else:
            return "region"

    @staticmethod
    def _get_map_cls(map_type):
        """Get map class for a given `map_type` string.

        This should probably be a registry dict so that users
        can add supported map types to the `gammapy.maps` I/O
        (see e.g. the Astropy table format I/O registry),
        but that's non-trivial to implement without avoiding circular imports.
        """
        if map_type == "wcs":
            from .wcs import WcsNDMap

            return WcsNDMap
        elif map_type == "wcs-sparse":
            raise NotImplementedError()
        elif map_type == "hpx":
            from .hpx import HpxNDMap

            return HpxNDMap
        elif map_type == "hpx-sparse":
            raise NotImplementedError()
        elif map_type == "region":
            from .region import RegionNDMap

            return RegionNDMap
        else:
            raise ValueError(f"Unrecognized map type: {map_type!r}")

    def write(self, filename, overwrite=False, **kwargs):
        """Write to a FITS file.

        Parameters
        ----------
        filename : str
            Output file name.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        hdu : str, optional
            Set the name of the image extension. By default, this will
            be set to 'SKYMAP' (for BINTABLE HDU) or 'PRIMARY' (for IMAGE
            HDU).
        hdu_bands : str, optional
            Set the name of the bands table extension. Default is 'BANDS'.
        format : str, optional
            FITS format convention. By default, files will be written
            to the gamma-astro-data-formats (GADF) format.  This
            option can be used to write files that are compliant with
            format conventions required by specific software (e.g. the
            Fermi Science Tools). The following formats are supported:

                - "gadf" (default)
                - "fgst-ccube"
                - "fgst-ltcube"
                - "fgst-bexpcube"
                - "fgst-srcmap"
                - "fgst-template"
                - "fgst-srcmap-sparse"
                - "galprop"
                - "galprop2"

        sparse : bool, optional
            Sparsify the map by dropping pixels with zero amplitude.
            This option is only compatible with the 'gadf' format.
        checksum : bool, optional
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        checksum = kwargs.pop("checksum", False)
        hdulist = self.to_hdulist(**kwargs)
        hdulist.writeto(make_path(filename), overwrite=overwrite, checksum=checksum)

    def iter_by_axis(self, axis_name, keepdims=False):
        """Iterate over a given axis.

        Yields
        ------
        map : `Map`
            Map iteration.

        See Also
        --------
        iter_by_image : iterate by image returning a map.
        """
        axis = self.geom.axes[axis_name]
        for idx in range(axis.nbin):
            idx_axis = slice(idx, idx + 1) if keepdims else idx
            slices = {axis_name: idx_axis}
            yield self.slice_by_idx(slices=slices)

    def iter_by_image(self, keepdims=False):
        """Iterate over image planes of a map.

        Parameters
        ----------
        keepdims : bool, optional
            Keep dimensions. Default is False.

        Yields
        ------
        map : `Map`
            Map iteration.

        See Also
        --------
        iter_by_image_data : iterate by image returning data and index.
        """
        for idx in np.ndindex(self.geom.shape_axes):
            if keepdims:
                names = self.geom.axes.names
                slices = {name: slice(_, _ + 1) for name, _ in zip(names, idx)}
                yield self.slice_by_idx(slices=slices)
            else:
                yield self.get_image_by_idx(idx=idx)

    def iter_by_image_data(self):
        """Iterate over image planes of the map.

        The image plane index is in data order, so that the data array can be
        indexed directly.

        Yields
        ------
        (data, idx) : tuple
            Where ``data`` is a `numpy.ndarray` view of the image plane data,
            and ``idx`` is a tuple of int, the index of the image plane.

        See Also
        --------
        iter_by_image : iterate by image returning a map.
        """
        for idx in np.ndindex(self.geom.shape_axes):
            yield self.data[idx[::-1]], idx[::-1]

    def iter_by_image_index(self):
        """Iterate over image planes of the map.

        The image plane index is in data order, so that the data array can be
        indexed directly.

        Yields
        ------
        idx : tuple
            ``idx`` is a tuple of int, the index of the image plane.

        See Also
        --------
        iter_by_image : iterate by image returning a map.
        """
        for idx in np.ndindex(self.geom.shape_axes):
            yield idx[::-1]

    def coadd(self, map_in, weights=None):
        """Add the contents of ``map_in`` to this map.

        This method can be used to combine maps containing integral quantities (e.g. counts)
        or differential quantities if the maps have the same binning.

        Parameters
        ----------
        map_in : `Map`
            Map to add.
        weights: `Map` or `~numpy.ndarray`
            The weight factors while adding. Default is None.
        """
        if not self.unit.is_equivalent(map_in.unit):
            raise ValueError("Incompatible units")

        # TODO: Check whether geometries are aligned and if so sum the
        # data vectors directly
        if weights is not None:
            map_in = map_in * weights
        idx = map_in.geom.get_idx()
        coords = map_in.geom.get_coord()
        vals = u.Quantity(map_in.get_by_idx(idx), map_in.unit)
        self.fill_by_coord(coords, vals)

    def pad(self, pad_width, axis_name=None, mode="constant", cval=0, method="linear"):
        """Pad the spatial dimensions of the map.

        Parameters
        ----------
        pad_width : {sequence, array_like, int}
            Number of pixels padded to the edges of each axis.
        axis_name : str, optional
            Which axis to downsample. By default, spatial axes are padded. Default is None.
        mode : {'constant', 'edge', 'interp'}
            Padding mode.  'edge' pads with the closest edge value.
            'constant' pads with a constant value. 'interp' pads with
            an extrapolated value. Default is 'constant'.
        cval : float, optional
            Padding value when ``mode='consant'``. Default is 0.

        Returns
        -------
        map : `Map`
            Padded map.

        """
        if axis_name:
            if np.isscalar(pad_width):
                pad_width = (pad_width, pad_width)

            geom = self.geom.pad(pad_width=pad_width, axis_name=axis_name)
            idx = self.geom.axes.index_data(axis_name)
            pad_width_np = [(0, 0)] * self.data.ndim
            pad_width_np[idx] = pad_width

            kwargs = {}
            if mode == "constant":
                kwargs["constant_values"] = cval

            data = np.pad(self.data, pad_width=pad_width_np, mode=mode, **kwargs)
            return self.__class__(
                geom=geom, data=data, unit=self.unit, meta=self.meta.copy()
            )

        return self._pad_spatial(pad_width, mode="constant", cval=cval)

    @abc.abstractmethod
    def _pad_spatial(self, pad_width, mode="constant", cval=0, order=1):
        pass

    @abc.abstractmethod
    def crop(self, crop_width):
        """Crop the spatial dimensions of the map.

        Parameters
        ----------
        crop_width : {sequence, array_like, int}
            Number of pixels cropped from the edges of each axis.
            Defined analogously to ``pad_with`` from `numpy.pad`.

        Returns
        -------
        map : `Map`
            Cropped map.
        """
        pass

    @abc.abstractmethod
    def downsample(self, factor, preserve_counts=True, axis_name=None):
        """Downsample the spatial dimension by a given factor.

        Parameters
        ----------
        factor : int
            Downsampling factor.
        preserve_counts : bool, optional
            Preserve the integral over each bin. This should be set to True
            if the map is an integral quantity (e.g. counts) and False if
            the map is a differential quantity (e.g. intensity). Default is True.
        axis_name : str, optional
            Which axis to downsample. By default, spatial axes are downsampled. Default is None.

        Returns
        -------
        map : `Map`
            Downsampled map.
        """
        pass

    @abc.abstractmethod
    def upsample(self, factor, order=0, preserve_counts=True, axis_name=None):
        """Upsample the spatial dimension by a given factor.

        Parameters
        ----------
        factor : int
            Upsampling factor.
        order : int, optional
            Order of the interpolation used for upsampling. Default is 0.
        preserve_counts : bool, optional
            Preserve the integral over each bin. This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity). Default is True.
        axis_name : str, optional
            Which axis to upsample. By default, spatial axes are upsampled. Default is None.


        Returns
        -------
        map : `Map`
            Upsampled map.
        """
        pass

    def resample(self, geom, weights=None, preserve_counts=True, smooth=False):
        """Resample pixels to ``geom`` with given ``weights``.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Target Map geometry.
        weights : `~numpy.ndarray`, optional
            Weights vector. It must have same shape as the data of the map.
            If set to None, weights will be set to 1. Default is None.
        preserve_counts : bool, optional
            Preserve the integral over each bin.  This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity). Default is True.

        Returns
        -------
        resampled_map : `Map`
            Resampled map.
        """
        coords = self.geom.get_coord()
        idx = geom.coord_to_idx(coords)

        weights = 1 if weights is None else weights

        resampled = self._init_copy(data=None, geom=geom)
        resampled._resample_by_idx(
            idx,
            weights=self.data * weights,
            preserve_counts=preserve_counts,
            smooth=smooth,
        )
        return resampled

    @abc.abstractmethod
    def _resample_by_idx(self, idx, weights=None, preserve_counts=False, smooth=False):
        """Resample pixels at ``idx`` with given ``weights``.

        Parameters
        ----------
        idx : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
        weights : `~numpy.ndarray`, optional
            Weights vector. If set to None, weights will be set to 1. Default is None.
        preserve_counts : bool, optional
            Preserve the integral over each bin.  This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity). Default is False.
        smooth : bool, optional
            Smooth output before scaling the total counts.
            Used only for WCS maps if preserve_counts is True. Default is False.

        """
        pass

    def resample_axis(self, axis, weights=None, ufunc=np.add):
        """Resample map to a new axis by grouping and reducing smaller bins by a given function `ufunc`.

        By default, the map content are summed over the smaller bins. Other `numpy.ufunc` can be
        used, e.g. `numpy.logical_and` or `numpy.logical_or`.

        Parameters
        ----------
        axis : `MapAxis`
            New map axis.
        weights : `Map`, optional
            Array to be used as weights. The spatial geometry must be equivalent
            to `other` and additional axes must be broadcastable. If set to None, weights will be set to 1.
            Default is None.
        ufunc : `~numpy.ufunc`
            Universal function to use to resample the axis. Default is `numpy.add`.

        Returns
        -------
        map : `Map`
            Map with resampled axis.
        """
        from .hpx import HpxGeom

        geom = self.geom.resample_axis(axis)

        axis_self = self.geom.axes[axis.name]
        axis_resampled = geom.axes[axis.name]

        # We don't use MapAxis.coord_to_idx because is does not behave as needed with boundaries
        coord = axis_resampled.edges.value
        edges = axis_self.edges.value
        indices = np.digitize(coord, edges) - 1

        idx = self.geom.axes.index_data(axis.name)

        weights = 1 if weights is None else weights.data

        if not isinstance(self.geom, HpxGeom):
            shape = self.geom._shape[:2]
        else:
            shape = (self.geom.data_shape[-1],)
        shape += tuple([ax.nbin if ax != axis else 1 for ax in self.geom.axes])

        padded_array = np.append(self.data * weights, np.zeros(shape[::-1]), axis=idx)

        slices = tuple([slice(0, _) for _ in geom.data_shape])
        data = ufunc.reduceat(padded_array, indices=indices, axis=idx)[slices]

        return self._init_copy(data=data, geom=geom)

    def slice_by_idx(
        self,
        slices,
    ):
        """Slice sub map from map object.

        Parameters
        ----------
        slices : dict
            Dictionary of axes names and integers or `slice` object pairs. Contains one
            element for each non-spatial dimension. For integer indexing the
            corresponding axes is dropped from the map. Axes not specified in the
            dictionary are kept unchanged.

        Returns
        -------
        map_out : `Map`
            Sliced map object.

        Examples
        --------
        >>> from gammapy.maps import Map
        >>> m = Map.read("$GAMMAPY_DATA/fermi_3fhl/gll_iem_v06_cutout.fits")
        >>> slices = {"energy": slice(0, 5)}
        >>> sliced = m.slice_by_idx(slices)
        """
        geom = self.geom.slice_by_idx(slices)
        slices = tuple([slices.get(ax.name, slice(None)) for ax in self.geom.axes])
        data = self.data[slices[::-1]]
        return self.__class__(geom=geom, data=data, unit=self.unit, meta=self.meta)

    def get_image_by_coord(self, coords):
        """Return spatial map at the given axis coordinates.

        Parameters
        ----------
        coords : tuple or dict
            Tuple should be ordered as (x_0, ..., x_n) where x_i are coordinates
            for non-spatial dimensions of the map. Dictionary should specify the axis
            names of the non-spatial axes such as {'axes0': x_0, ..., 'axesn': x_n}.

        Returns
        -------
        map_out : `Map`
            Map with spatial dimensions only.

        See Also
        --------
        get_image_by_idx, get_image_by_pix.

        Examples
        --------
        ::

            import numpy as np
            from gammapy.maps import Map, MapAxis
            from astropy.coordinates import SkyCoord
            from astropy import units as u

            # Define map axes
            energy_axis = MapAxis.from_edges(
                np.logspace(-1., 1., 4), unit='TeV', name='energy',
            )

            time_axis = MapAxis.from_edges(
                np.linspace(0., 10, 20), unit='h', name='time',
            )

            # Define map center
            skydir = SkyCoord(0, 0, frame='galactic', unit='deg')

            # Create map
            m_wcs = Map.create(
                map_type='wcs',
                binsz=0.02,
                skydir=skydir,
                width=10.0,
                axes=[energy_axis, time_axis],
            )

            # Get image by coord tuple
            image = m_wcs.get_image_by_coord(('500 GeV', '1 h'))

            # Get image by coord dict with strings
            image = m_wcs.get_image_by_coord({'energy': '500 GeV', 'time': '1 h'})

            # Get image by coord dict with quantities
            image = m_wcs.get_image_by_coord({'energy': 0.5 * u.TeV, 'time': 1 * u.h})
        """
        if isinstance(coords, tuple):
            coords = dict(zip(self.geom.axes.names, coords))

        idx = self.geom.axes.coord_to_idx(coords)
        return self.get_image_by_idx(idx)

    def get_image_by_pix(self, pix):
        """Return spatial map at the given axis pixel coordinates.

        Parameters
        ----------
        pix : tuple
            Tuple of scalar pixel coordinates for each non-spatial dimension of
            the map. Tuple should be ordered as (I_0, ..., I_n). Pixel coordinates
            can be either float or integer type.

        See Also
        --------
        get_image_by_coord, get_image_by_idx.

        Returns
        -------
        map_out : `Map`
            Map with spatial dimensions only.
        """
        idx = self.geom.pix_to_idx(pix)
        return self.get_image_by_idx(idx)

    def get_image_by_idx(self, idx):
        """Return spatial map at the given axis pixel indices.

        Parameters
        ----------
        idx : tuple
            Tuple of scalar indices for each non-spatial dimension of the map.
            Tuple should be ordered as (I_0, ..., I_n).

        See Also
        --------
        get_image_by_coord, get_image_by_pix.

        Returns
        -------
        map_out : `Map`
            Map with spatial dimensions only.
        """
        if len(idx) != len(self.geom.axes):
            raise ValueError("Tuple length must equal number of non-spatial dimensions")

        # Only support scalar indices per axis
        idx = tuple([int(_.item()) for _ in np.array(idx)])

        geom = self.geom.to_image()
        data = self.data[idx[::-1]]
        return self.__class__(geom=geom, data=data, unit=self.unit, meta=self.meta)

    def get_by_coord(self, coords, fill_value=np.nan):
        """Return map values at the given map coordinates.

        Parameters
        ----------
        coords : tuple or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map. Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
        fill_value : float
            Value which is returned if the position is outside the projection
            footprint. Default is `numpy.nan`.

        Returns
        -------
        vals : `~numpy.ndarray`
           Values of pixels in the map. `numpy.nan` is used to flag coordinates
           outside the map.
        """
        pix = self.geom.coord_to_pix(coords=coords)
        vals = self.get_by_pix(pix, fill_value=fill_value)
        return vals

    def get_by_pix(self, pix, fill_value=np.nan):
        """Return map values at the given pixel coordinates.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
            Pixel indices can be either float or integer type.
        fill_value : float
            Value which is returned if the position is outside the projection
            footprint. Default is `numpy.nan`.

        Returns
        -------
        vals : `~numpy.ndarray`
           Array of pixel values. `numpy.nan` is used to flag coordinates
           outside the map.
        """
        # FIXME: Support local indexing here?
        # FIXME: Support slicing?
        pix = np.broadcast_arrays(*pix)
        idx = self.geom.pix_to_idx(pix)
        vals = self.get_by_idx(idx)
        mask = self.geom.contains_pix(pix)

        if not mask.all():
            vals = vals.astype(type(fill_value))
            vals[~mask] = fill_value

        return vals

    @abc.abstractmethod
    def get_by_idx(self, idx):
        """Return map values at the given pixel indices.

        Parameters
        ----------
        idx : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.

        Returns
        -------
        vals : `~numpy.ndarray`
           Array of pixel values. `numpy.nan` is used to flag coordinates outside the map.
        """
        pass

    @abc.abstractmethod
    def interp_by_coord(self, coords, method="linear", fill_value=None):
        """Interpolate map values at the given map coordinates.

        Parameters
        ----------
        coords : tuple or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map. Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
        method : {"linear", "nearest"}
            Method to interpolate data values. Default is "linear".
        fill_value : float, optional
            The value to use for points outside the interpolation domain.
            If None, values outside the domain are extrapolated. Default is None.

        Returns
        -------
        vals : `~numpy.ndarray`
            Interpolated pixel values.
        """
        pass

    @abc.abstractmethod
    def interp_by_pix(self, pix, method="linear", fill_value=None):
        """Interpolate map values at the given pixel coordinates.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel coordinate arrays for each dimension of the
            map. Tuple should be ordered as (p_lon, p_lat, p_0, ...,
            p_n) where p_i are pixel coordinates for non-spatial
            dimensions of the map.
        method : {"linear", "nearest"}
            Method to interpolate data values. Default is "linear".
        fill_value : float, optional
            The value to use for points outside the interpolation domain.
            If None, values outside the domain are extrapolated. Default is None.

        Returns
        -------
        vals : `~numpy.ndarray`
            Interpolated pixel values.
        """
        pass

    def interp_to_geom(self, geom, preserve_counts=False, fill_value=0, **kwargs):
        """Interpolate map to input geometry.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Target Map geometry.
        preserve_counts : bool, optional
            Preserve the integral over each bin. This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity). Default is False.
        fill_value : float, optional
            The value to use for points outside the interpolation domain.
            If None, values outside the domain are extrapolated.
            Default is 0.
        **kwargs : dict, optional
            Keyword arguments passed to `Map.interp_by_coord`.

        Returns
        -------
        interp_map : `Map`
            Interpolated Map.
        """
        coords = geom.get_coord()
        map_copy = self.copy()

        if preserve_counts:
            if geom.ndim > 2 and geom.axes[0] != self.geom.axes[0]:
                raise ValueError(
                    f"Energy axes do not match, expected: \n {self.geom.axes[0]},"
                    f" but got: \n {geom.axes[0]}."
                )
            map_copy.data /= map_copy.geom.solid_angle().to_value("deg2")

        if map_copy.is_mask and fill_value is not None:
            # TODO: check this NaN handling is needed
            data = map_copy.get_by_coord(coords)
            data = np.nan_to_num(data, nan=fill_value).astype(bool)
        else:
            data = map_copy.interp_by_coord(coords, fill_value=fill_value, **kwargs)
            if map_copy.is_mask:
                data = data.astype(bool)

        if preserve_counts:
            data *= geom.solid_angle().to_value("deg2")

        return Map.from_geom(geom, data=data, unit=self.unit)

    def reproject_to_geom(self, geom, preserve_counts=False, precision_factor=10):
        """Reproject map to input geometry.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Target Map geometry.
        preserve_counts : bool, optional
            Preserve the integral over each bin. This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity). Default is False.
        precision_factor : int, optional
           Minimal factor between the bin size of the output map and the oversampled base map.
           Used only for the oversampling method. Default is 10.

        Returns
        -------
        output_map : `Map`
            Reprojected Map.
        """
        from .hpx import HpxGeom
        from .region import RegionGeom

        axes = [ax.copy() for ax in self.geom.axes]
        geom3d = geom.copy(axes=axes)

        if not geom.is_image:
            if geom.axes.names != geom3d.axes.names:
                raise ValueError("Axis names and order should be the same.")
            if geom.axes != geom3d.axes and (
                isinstance(geom3d, HpxGeom) or isinstance(self.geom, HpxGeom)
            ):
                raise TypeError(
                    "Reprojection to 3d geom with non-identical axes is not supported for HpxGeom. "
                    "Reproject to 2d geom first and then use inter_to_geom method."
                )
        if isinstance(geom3d, RegionGeom):
            base_factor = (
                geom3d.to_wcs_geom().pixel_scales.min() / self.geom.pixel_scales.min()
            )
        elif isinstance(self.geom, RegionGeom):
            base_factor = (
                geom3d.pixel_scales.min() / self.geom.to_wcs_geom().pixel_scales.min()
            )
        else:
            base_factor = geom3d.pixel_scales.min() / self.geom.pixel_scales.min()

        if base_factor >= precision_factor:
            input_map = self
        else:
            factor = precision_factor / base_factor
            if isinstance(self.geom, HpxGeom):
                factor = int(2 ** np.ceil(np.log(factor) / np.log(2)))
            else:
                factor = int(np.ceil(factor))
            input_map = self.upsample(factor=factor, preserve_counts=preserve_counts)

        output_map = input_map.resample(
            geom3d, preserve_counts=preserve_counts, smooth=True
        )

        if not geom.is_image and geom.axes != geom3d.axes:
            for base_ax, target_ax in zip(geom3d.axes, geom.axes):
                base_factor = base_ax.bin_width.min() / target_ax.bin_width.min()
                if not base_factor >= precision_factor:
                    factor = precision_factor / base_factor
                    factor = int(np.ceil(factor))
                    output_map = output_map.upsample(
                        factor=factor,
                        preserve_counts=preserve_counts,
                        axis_name=base_ax.name,
                    )
            output_map = output_map.resample(
                geom, preserve_counts=preserve_counts, smooth=True
            )
        return output_map

    def reproject_by_image(
        self,
        geom,
        preserve_counts=False,
        precision_factor=10,
    ):
        """Reproject each image of a ND map to input 2d geometry.

        For large maps this method is faster than `reproject_to_geom`.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Target slice geometry in 2D.
        preserve_counts : bool, optional
            Preserve the integral over each bin. This should be True
            if the map is an integral quantity (e.g. counts) and False if
            the map is a differential quantity (e.g. intensity). Default is False.
        precision_factor : int, optional
            Minimal factor between the bin size of the output map and the oversampled base map.
            Used only for the oversampling method. Default is 10.

        Returns
        -------
        output_map : `Map`
            Reprojected Map.
        """
        if not geom.is_image:
            raise TypeError("This method is only valid for 2d geom")

        output_map = Map.from_geom(geom.to_cube(self.geom.axes))
        maps = parallel.run_multiprocessing(
            self._reproject_image,
            zip(
                self.iter_by_image(),
                repeat(geom),
                repeat(preserve_counts),
                repeat(precision_factor),
            ),
            task_name="Reprojection",
        )
        for idx in np.ndindex(self.geom.shape_axes):
            output_map.data[idx[0]] = maps[idx[0]].data
        return output_map

    @staticmethod
    def _reproject_image(image, geom, preserve_counts, precision_factor):
        return image.reproject_to_geom(
            geom, precision_factor=precision_factor, preserve_counts=preserve_counts
        )

    def fill_events(self, events, weights=None):
        """Fill the map from an `~gammapy.data.EventList` object.

        Parameters
        ----------
        events : `~gammapy.data.EventList`
            Events to fill in the map with.
        weights : `~numpy.ndarray`, optional
            Weights vector. The weights vector must be of the same length
            as the events column length. If None, weights are set to 1. Default is None.
        """
        self.fill_by_coord(events.map_coord(self.geom), weights=weights)

    def fill_by_coord(self, coords, weights=None):
        """Fill pixels at ``coords`` with given ``weights``.

        Parameters
        ----------
        coords : tuple or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map. Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
        weights : `~numpy.ndarray`, optional
            Weights vector. If None, weights are set to 1. Default is None.
        """
        idx = self.geom.coord_to_idx(coords)
        self.fill_by_idx(idx, weights=weights)

    def fill_by_pix(self, pix, weights=None):
        """Fill pixels at ``pix`` with given ``weights``.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
            Pixel indices can be either float or integer type. Float
            indices will be rounded to the nearest integer.
        weights : `~numpy.ndarray`, optional
            Weights vector. If None, weights are set to 1. Default is None.
        """
        idx = pix_tuple_to_idx(pix)
        return self.fill_by_idx(idx, weights=weights)

    @abc.abstractmethod
    def fill_by_idx(self, idx, weights=None):
        """Fill pixels at ``idx`` with given ``weights``.

        Parameters
        ----------
        idx : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
        weights : `~numpy.ndarray`, optional
            Weights vector. If None, weights are set to 1. Default is None.
        """
        pass

    def set_by_coord(self, coords, vals):
        """Set pixels at ``coords`` with given ``vals``.

        Parameters
        ----------
        coords : tuple or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map. Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
        vals : `~numpy.ndarray`
            Values vector.
        """
        idx = self.geom.coord_to_pix(coords)
        self.set_by_pix(idx, vals)

    def set_by_pix(self, pix, vals):
        """Set pixels at ``pix`` with given ``vals``.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
            Pixel indices can be either float or integer type. Float
            indices will be rounded to the nearest integer.
        vals : `~numpy.ndarray`
            Values vector.
        """
        idx = pix_tuple_to_idx(pix)
        return self.set_by_idx(idx, vals)

    @abc.abstractmethod
    def set_by_idx(self, idx, vals):
        """Set pixels at ``idx`` with given ``vals``.

        Parameters
        ----------
        idx : tuple
            Tuple of pixel index arrays for each dimension of the map.
            Tuple should be ordered as (I_lon, I_lat, I_0, ..., I_n)
            for WCS maps and (I_hpx, I_0, ..., I_n) for HEALPix maps.
        vals : `~numpy.ndarray`
            Values vector.
        """
        pass

    def plot_grid(self, figsize=None, ncols=3, **kwargs):
        """Plot map as a grid of subplots for non-spatial axes.

        Parameters
        ----------
        figsize : tuple of int, optional
            Figsize to plot on. Default is None.
        ncols : int, optional
            Number of columns to plot. Default is 3.
        **kwargs : dict, optional
            Keyword arguments passed to `WcsNDMap.plot`.

        Returns
        -------
        axes : `~numpy.ndarray` of `~matplotlib.pyplot.Axes`
            Axes grid.
        """
        if len(self.geom.axes) > 1:
            raise ValueError("Grid plotting is only supported for one non spatial axis")

        axis = self.geom.axes[0]

        cols = min(ncols, axis.nbin)
        rows = 1 + (axis.nbin - 1) // cols

        if figsize is None:
            width = 12
            figsize = (width, width * rows / cols)

        if self.geom.is_hpx:
            wcs = self.geom.to_wcs_geom().wcs
        else:
            wcs = self.geom.wcs

        fig, axes = plt.subplots(
            ncols=cols,
            nrows=rows,
            subplot_kw={"projection": wcs},
            figsize=figsize,
            gridspec_kw={"hspace": 0.1, "wspace": 0.1},
        )

        for idx in range(cols * rows):
            ax = axes.flat[idx]

            try:
                image = self.get_image_by_idx((idx,))
            except IndexError:
                ax.set_visible(False)
                continue

            if image.geom.is_hpx:
                image_wcs = image.to_wcs(
                    normalize=False,
                    proj="AIT",
                    oversample=2,
                )
            else:
                image_wcs = image

            image_wcs.plot(ax=ax, **kwargs)

            if axis.node_type == "center":
                if axis.name == "energy" or axis.name == "energy_true":
                    info = energy_unit_format(axis.center[idx])
                else:
                    info = f"{axis.center[idx]:.1f}"
            elif axis.node_type == "label":
                info = f"{axis.center[idx]}"
            else:
                if axis.name == "energy" or axis.name == "energy_true":
                    info = (
                        f"{energy_unit_format(axis.edges[idx])} - "
                        f"{energy_unit_format(axis.edges[idx+1])}"
                    )
                else:
                    info = f"{axis.edges_min[idx]:.1f} - {axis.edges_max[idx]:.1f} "
            ax.set_title(f"{axis.name.capitalize()} " + info)
            lon, lat = ax.coords[0], ax.coords[1]
            lon.set_ticks_position("b")
            lat.set_ticks_position("l")

            row, col = np.unravel_index(idx, shape=(rows, cols))

            if col > 0:
                lat.set_ticklabel_visible(False)
                lat.set_axislabel("")

            if row < (rows - 1):
                lon.set_ticklabel_visible(False)
                lon.set_axislabel("")

        return axes

    def plot_interactive(self, rc_params=None, **kwargs):
        """
        Plot map with interactive widgets to explore the non-spatial axes.

        Parameters
        ----------
        rc_params : dict, optional
            Passed to ``matplotlib.rc_context(rc=rc_params)`` to style the plot. Default is None.
        **kwargs : dict, optional
            Keyword arguments passed to `WcsNDMap.plot`.

        Examples
        --------
        You can try this out e.g. using a Fermi-LAT diffuse model cube with an energy axis::

            from gammapy.maps import Map

            m = Map.read("$GAMMAPY_DATA/fermi_3fhl/gll_iem_v06_cutout.fits")
            m.plot_interactive(add_cbar=True, stretch="sqrt")

        If you would like to adjust the figure size you can use the ``rc_params`` argument::

            rc_params = {'figure.figsize': (12, 6), 'font.size': 12}
            m.plot_interactive(rc_params=rc_params)
        """
        import matplotlib as mpl
        from ipywidgets import RadioButtons, SelectionSlider
        from ipywidgets.widgets.interaction import fixed, interact

        if self.geom.is_image:
            raise TypeError("Use .plot() for 2D Maps")

        kwargs.setdefault("interpolation", "nearest")
        kwargs.setdefault("origin", "lower")
        kwargs.setdefault("cmap", "afmhot")

        rc_params = rc_params or {}
        stretch = kwargs.pop("stretch", "sqrt")

        interact_kwargs = {}

        for axis in self.geom.axes:
            if axis.node_type == "center":
                if axis.name == "energy" or axis.name == "energy_true":
                    options = energy_unit_format(axis.center)
                else:
                    options = axis.as_plot_labels
            elif axis.name == "energy" or axis.name == "energy_true":
                E = energy_unit_format(axis.edges)
                options = [f"{E[i]} - {E[i+1]}" for i in range(len(E) - 1)]
            else:
                options = axis.as_plot_labels
            interact_kwargs[axis.name] = SelectionSlider(
                options=options,
                description=f"Select {axis.name}:",
                continuous_update=False,
                style={"description_width": "initial"},
                layout={"width": "50%"},
            )
            interact_kwargs[axis.name + "_options"] = fixed(options)

        interact_kwargs["stretch"] = RadioButtons(
            options=["linear", "sqrt", "log"],
            value=stretch,
            description="Select stretch:",
            style={"description_width": "initial"},
        )

        @interact(**interact_kwargs)
        def _plot_interactive(**ikwargs):
            idx = [
                ikwargs[ax.name + "_options"].index(ikwargs[ax.name])
                for ax in self.geom.axes
            ]
            img = self.get_image_by_idx(idx)
            stretch = ikwargs["stretch"]
            with mpl.rc_context(rc=rc_params):
                img.plot(stretch=stretch, **kwargs)
                plt.show()

    def copy(self, **kwargs):
        """Copy map instance and overwrite given attributes, except for geometry.

        Parameters
        ----------
        **kwargs : dict, optional
            Keyword arguments to overwrite in the map constructor.

        Returns
        -------
        copy : `Map`
            Copied Map.
        """
        if "geom" in kwargs:
            geom = kwargs["geom"]
            if not geom.data_shape == self.geom.data_shape:
                raise ValueError(
                    "Can't copy and change data size of the map. "
                    f" Current shape {self.geom.data_shape},"
                    f" requested shape {geom.data_shape}"
                )

        return self._init_copy(**kwargs)

    def mask_nearest_position(self, position):
        """Given a sky coordinate return nearest valid position in the mask.

        If the mask contains additional axes, the mask is reduced over those axes.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Test position.

        Returns
        -------
        position : `~astropy.coordinates.SkyCoord`
            The nearest position in the mask.
        """
        if not self.geom.is_image:
            raise ValueError("Method only supported for 2D images")

        coords = self.geom.to_image().get_coord().skycoord
        separation = coords.separation(position)
        separation[~self.data] = np.inf
        idx = np.argmin(separation)
        return coords.flatten()[idx]

    def sum_over_axes(self, axes_names=None, keepdims=True, weights=None):
        """Sum map values over all non-spatial axes.

        Parameters
        ----------
         axes_names: list of str
            Names of the axis to reduce over. If None, all non-spatial axis will be summed over. Default is None.
        keepdims : bool, optional
            If this is set to true, the axes which are summed over are left in
            the map with a single bin. Default is True.
        weights : `Map`, optional
            Weights to be applied. The map should have the same geometry as this one. Default is None.

        Returns
        -------
        map_out : `~Map`
            Map with non-spatial axes summed over.
        """
        return self.reduce_over_axes(
            func=np.add, axes_names=axes_names, keepdims=keepdims, weights=weights
        )

    def reduce_over_axes(
        self, func=np.add, keepdims=False, axes_names=None, weights=None
    ):
        """Reduce map over non-spatial axes.

        Parameters
        ----------
        func : `~numpy.ufunc`, optional
            Function to use for reducing the data. Default is `numpy.add`.
        keepdims : bool, optional
            If this is set to true, the axes which are summed over are left in
            the map with a single bin. Default is False.
        axes_names: list, optional
            Names of axis to reduce over. If None, all non-spatial axis will be reduced.
        weights : `Map`, optional
            Weights to be applied. The map should have the same geometry as this one. Default is None.

        Returns
        -------
        map_out : `~Map`
            Map with non-spatial axes reduced.
        """
        if axes_names is None:
            axes_names = self.geom.axes.names

        map_out = self.copy()
        for axis_name in axes_names:
            map_out = map_out.reduce(
                axis_name, func=func, keepdims=keepdims, weights=weights
            )

        return map_out

    def reduce(self, axis_name, func=np.add, keepdims=False, weights=None):
        """Reduce map over a single non-spatial axis.

        Parameters
        ----------
        axis_name: str
            The name of the axis to reduce over.
        func : `~numpy.ufunc`, optional
            Function to use for reducing the data. Default is `numpy.add`.
        keepdims : bool, optional
            If this is set to true, the axes which are summed over are left in
            the map with a single bin. Default is False.
        weights : `Map`
            Weights to be applied. The map should have the same geometry as this one. Default is None.

        Returns
        -------
        map_out : `~Map`
            Map with the given non-spatial axes reduced.
        """
        if keepdims:
            geom = self.geom.squash(axis_name=axis_name)
        else:
            geom = self.geom.drop(axis_name=axis_name)

        idx = self.geom.axes.index_data(axis_name)

        data = self.data

        if weights is not None:
            data = data * weights

        data = func.reduce(data, axis=idx, keepdims=keepdims, where=~np.isnan(data))
        return self._init_copy(geom=geom, data=data)

    def cumsum(self, axis_name):
        """Compute cumulative sum along a given axis.

        Parameters
        ----------
        axis_name : str
            Along which axis to sum.

        Returns
        -------
        cumsum : `Map`
            Map with cumulative sum.
        """
        axis = self.geom.axes[axis_name]
        axis_idx = self.geom.axes.index_data(axis_name)

        # TODO: the broadcasting should be done by axis.center, axis.bin_width etc.
        shape = [1] * len(self.geom.data_shape)
        shape[axis_idx] = -1

        values = self.quantity * axis.bin_width.reshape(shape)

        if axis_name == "rad":
            # take Jacobian into account
            values = 2 * np.pi * axis.center.reshape(shape) * values

        data = np.insert(values.cumsum(axis=axis_idx), 0, 0, axis=axis_idx)

        axis_shifted = MapAxis.from_nodes(
            axis.edges, name=axis.name, interp=axis.interp
        )
        axes = self.geom.axes.replace(axis_shifted)
        geom = self.geom.to_image().to_cube(axes)
        return self.__class__(geom=geom, data=data.value, unit=data.unit)

    def integral(self, axis_name, coords, **kwargs):
        """Compute integral along a given axis.

        This method uses interpolation of the cumulative sum.

        Parameters
        ----------
        axis_name : str
            Along which axis to integrate.
        coords : dict or `MapCoord`
            Map coordinates.
        **kwargs : dict, optional
            Keyword arguments passed to `Map.interp_by_coord`.

        Returns
        -------
        array : `~astropy.units.Quantity`
            2D array with axes offset.
        """
        cumsum = self.cumsum(axis_name=axis_name)
        cumsum = cumsum.pad(pad_width=1, axis_name=axis_name, mode="edge")
        return u.Quantity(
            cumsum.interp_by_coord(coords, **kwargs), cumsum.unit, copy=COPY_IF_NEEDED
        )

    def normalize(self, axis_name=None):
        """Normalise data in place along a given axis.

        Parameters
        ----------
        axis_name : str, optional
            Along which axis to normalise.
        """
        cumsum = self.cumsum(axis_name=axis_name).quantity

        with np.errstate(invalid="ignore", divide="ignore"):
            axis = self.geom.axes.index_data(axis_name=axis_name)
            normed = self.quantity / cumsum.max(axis=axis, keepdims=True)

        self.quantity = np.nan_to_num(normed)

    @classmethod
    def from_stack(cls, maps, axis=None, axis_name=None):
        """Create Map from a list of images and a non-spatial axis.

        The image geometries must be aligned, except for the axis that is stacked.

        Parameters
        ----------
        maps : list of `Map` objects
            List of maps.
        axis : `MapAxis`, optional
            If a `MapAxis` is provided the maps are stacked along the last data
            axis and the new axis is introduced. Default is None.
        axis_name : str, optional
            If an axis name is as string the given the maps are stacked along
            the given axis name.

        Returns
        -------
        map : `Map`
            Map with additional non-spatial axis.
        """
        geom = maps[0].geom

        if axis_name is None and axis is None:
            axis_name = geom.axes.names[-1]

        if axis_name:
            axis = MapAxis.from_stack(axes=[m.geom.axes[axis_name] for m in maps])
            geom = geom.drop(axis_name=axis_name)

        data = []

        for m in maps:
            if axis_name:
                m_geom = m.geom.drop(axis_name=axis_name)
            else:
                m_geom = m.geom

            if not m_geom == geom:
                raise ValueError(f"Image geometries not aligned: {m.geom} and {geom}")

            data.append(m.quantity.to_value(maps[0].unit))

        new_geom = geom.to_cube(axes=[axis])
        data = np.concatenate(data).reshape(new_geom.data_shape)

        return cls.from_geom(data=data, geom=new_geom, unit=maps[0].unit)

    def split_by_axis(self, axis_name):
        """Split a Map along an axis into multiple maps.

        Parameters
        ----------
        axis_name : str
            Name of the axis to split.

        Returns
        -------
        maps : list
            A list of `~gammapy.maps.Map`.
        """
        maps = []
        axis = self.geom.axes[axis_name]
        for idx in range(axis.nbin):
            m = self.slice_by_idx({axis_name: idx})
            maps.append(m)
        return maps

    def to_cube(self, axes):
        """Append non-spatial axes to create a higher-dimensional Map.

        This will result in a Map with a new geometry with
        N+M dimensions where N is the number of current dimensions and
        M is the number of axes in the list. The data is reshaped onto this
        new geometry.

        Parameters
        ----------
        axes : list
            Axes that will be appended to this Map.
            The axes should have only one bin.

        Returns
        -------
        map : `~gammapy.maps.WcsNDMap`
            New map.
        """
        for ax in axes:
            if ax.nbin > 1:
                raise ValueError(ax.name, "should have only one bin")
        geom = self.geom.to_cube(axes)
        data = self.data.reshape((1,) * len(axes) + self.data.shape)
        return self.from_geom(data=data, geom=geom, unit=self.unit)

    def get_spectrum(self, region=None, func=np.nansum, weights=None):
        """Extract spectrum in a given region.

        The spectrum can be computed by summing (or, more generally, applying ``func``)
        along the spatial axes in each energy bin. This occurs only inside the ``region``,
        which by default is assumed to be the whole spatial extension of the map.

        Parameters
        ----------
        region: `~regions.Region`, optional
             Region to extract the spectrum from. Pixel or sky regions are accepted. Default is None.
        func : `numpy.func`, optional
            Function to reduce the data. Default is `~numpy.nansum`.
            For a boolean Map, use `numpy.any` or `numpy.all`. Default is `numpy.nansum`.
        weights : `WcsNDMap`, optional
            Array to be used as weights. The geometry must be equivalent. Default is None.

        Returns
        -------
        spectrum : `~gammapy.maps.RegionNDMap`
            Spectrum in the given region.
        """
        if not self.geom.has_energy_axis:
            raise ValueError("Energy axis required")

        return self.to_region_nd_map(region=region, func=func, weights=weights)

    def to_unit(self, unit):
        """Convert map to a different unit.

        Parameters
        ----------
        unit : str or `~astropy.unit.Unit`
            New unit.

        Returns
        -------
        map : `Map`
            Map with new unit and converted data.
        """
        data = self.quantity.to_value(unit)
        return self.from_geom(self.geom, data=data, unit=unit)

    def is_allclose(self, other, rtol_axes=1e-3, atol_axes=1e-6, **kwargs):
        """Compare two Maps for close equivalency.

        Parameters
        ----------
        other : `gammapy.maps.Map`
            The Map to compare against.
        rtol_axes : float, optional
            Relative tolerance for the axes' comparison. Default is 1e-3.
        atol_axes : float, optional
            Absolute tolerance for the axes' comparison. Default is 1e-6.
        **kwargs : dict, optional
                Keywords passed to `~numpy.allclose`.

        Returns
        -------
        is_allclose : bool
            Whether the Map is all close.
        """
        if not isinstance(other, self.__class__):
            return TypeError(f"Cannot compare {type(self)} and {type(other)}")

        if self.data.shape != other.data.shape:
            return False

        axes_eq = self.geom.axes.is_allclose(
            other.geom.axes, rtol=rtol_axes, atol=atol_axes
        )
        data_eq = np.allclose(self.quantity, other.quantity, **kwargs)
        return axes_eq and data_eq

    def __str__(self):
        geom = self.geom.__class__.__name__
        axes = ["skycoord"] if self.geom.is_hpx else ["lon", "lat"]
        axes = axes + [_.name for _ in self.geom.axes]

        return (
            f"{self.__class__.__name__}\n\n"
            f"\tgeom  : {geom} \n "
            f"\taxes  : {axes}\n"
            f"\tshape : {self.geom.data_shape[::-1]}\n"
            f"\tndim  : {self.geom.ndim}\n"
            f"\tunit  : {self.unit}\n"
            f"\tdtype : {self.data.dtype}\n"
        )

    def _arithmetics(self, operator, other, copy):
        """Perform arithmetic on maps after checking geometry consistency."""
        if isinstance(other, Map):
            if self.geom == other.geom:
                q = other.quantity
            else:
                raise ValueError("Map Arithmetic: Inconsistent geometries.")
        else:
            q = u.Quantity(other, copy=COPY_IF_NEEDED)

        out = self.copy() if copy else self
        out.quantity = operator(out.quantity, q)
        return out

    def _boolean_arithmetics(self, operator, other, copy):
        """Perform arithmetic on maps after checking geometry consistency."""
        if operator == np.logical_not:
            out = self.copy()
            out.data = operator(out.data)
            return out

        if isinstance(other, Map):
            if self.geom == other.geom:
                other = other.data
            else:
                raise ValueError("Map Arithmetic: Inconsistent geometries.")

        out = self.copy() if copy else self
        out.data = operator(out.data, other)
        return out

    def __add__(self, other):
        return self._arithmetics(np.add, other, copy=True)

    def __iadd__(self, other):
        return self._arithmetics(np.add, other, copy=COPY_IF_NEEDED)

    def __sub__(self, other):
        return self._arithmetics(np.subtract, other, copy=True)

    def __isub__(self, other):
        return self._arithmetics(np.subtract, other, copy=COPY_IF_NEEDED)

    def __mul__(self, other):
        return self._arithmetics(np.multiply, other, copy=True)

    def __imul__(self, other):
        return self._arithmetics(np.multiply, other, copy=COPY_IF_NEEDED)

    def __truediv__(self, other):
        return self._arithmetics(np.true_divide, other, copy=True)

    def __itruediv__(self, other):
        return self._arithmetics(np.true_divide, other, copy=COPY_IF_NEEDED)

    def __le__(self, other):
        return self._arithmetics(np.less_equal, other, copy=True)

    def __lt__(self, other):
        return self._arithmetics(np.less, other, copy=True)

    def __ge__(self, other):
        return self._arithmetics(np.greater_equal, other, copy=True)

    def __gt__(self, other):
        return self._arithmetics(np.greater, other, copy=True)

    def __eq__(self, other):
        return self._arithmetics(np.equal, other, copy=True)

    def __ne__(self, other):
        return self._arithmetics(np.not_equal, other, copy=True)

    def __and__(self, other):
        return self._boolean_arithmetics(np.logical_and, other, copy=True)

    def __or__(self, other):
        return self._boolean_arithmetics(np.logical_or, other, copy=True)

    def __invert__(self):
        return self._boolean_arithmetics(np.logical_not, None, copy=True)

    def __xor__(self, other):
        return self._boolean_arithmetics(np.logical_xor, other, copy=True)

    def __iand__(self, other):
        return self._boolean_arithmetics(np.logical_and, other, copy=COPY_IF_NEEDED)

    def __ior__(self, other):
        return self._boolean_arithmetics(np.logical_or, other, copy=COPY_IF_NEEDED)

    def __ixor__(self, other):
        return self._boolean_arithmetics(np.logical_xor, other, copy=COPY_IF_NEEDED)

    def __array__(self):
        return self.data

    def sample_coord(self, n_events, random_state=0):
        """Sample position and energy of events.

        Parameters
        ----------
        n_events : int
            Number of events to sample.
        random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}
            Defines random number generator initialisation.
            Passed to `~gammapy.utils.random.get_random_state`. Default is 0.

        Returns
        -------
        coords : `~gammapy.maps.MapCoord` object.
            Sequence of coordinates and energies of the sampled events.
        """
        random_state = get_random_state(random_state)
        sampler = InverseCDFSampler(pdf=self.data, random_state=random_state)

        coords_pix = sampler.sample(n_events)
        coords = self.geom.pix_to_coord(coords_pix[::-1])

        # TODO: pix_to_coord should return a MapCoord object
        cdict = OrderedDict(zip(self.geom.axes_names, coords))

        return MapCoord.create(cdict, frame=self.geom.frame)

    def reorder_axes(self, axes_names):
        """Return a new map re-ordering the non-spatial axes.

        Parameters
        ----------
        axes_names : list of str
            The list of axes names in the required order.

        Returns
        -------
        map : `~gammapy.maps.Map`
            The map with axes re-ordered.
        """
        old_axes = self.geom.axes
        if not set(old_axes.names) == set(axes_names):
            raise ValueError(f"{old_axes.names} is not compatible with {axes_names}")

        new_axes = [old_axes[_] for _ in axes_names]
        new_geom = self.geom.to_image().to_cube(new_axes)

        old_indices = [old_axes.index_data(ax) for ax in axes_names]
        new_indices = [new_geom.axes.index_data(ax) for ax in axes_names]

        data = np.moveaxis(self.data, old_indices, new_indices)

        return Map.from_geom(new_geom, data=data)

    def dot(self, other):
        """Apply dot product with the input map.

        The input Map has to share a single MapAxis with the current Map.
        Because it has no spatial dimension, it must be a `~gammapy.maps.RegionNDMap`.

        Parameters
        ----------
        other : `~gammapy.maps.RegionNDMap`
            Map to apply the dot product to.
            It must share a unique non-spatial MapAxis with the current Map.

        Returns
        -------
        map : `~gammapy.maps.Map`
            Map with dot product applied.
        """
        from .region import RegionNDMap

        if not isinstance(other, RegionNDMap):
            raise TypeError(
                f"Dot product can be applied to a RegionNDMap. Got {type(other)} instead."
            )

        common_names = list(
            set(other.geom.axes.names).intersection(self.geom.axes.names)
        )

        if len(common_names) == 0:
            raise ValueError(
                "Map geometries have no axis in common. Cannot apply dot product."
            )
        elif len(common_names) > 1:
            raise ValueError(
                f"Map geometries have more than one axis in common: {common_names}."
                "Cannot apply dot product."
            )

        axis_name = common_names[0]

        if self.geom.axes[axis_name] != other.geom.axes[axis_name]:
            raise ValueError(
                f"Axes {axis_name} are not equal. Cannot apply dot product."
            )

        loc = self.geom.axes.index_data(axis_name)
        other_loc = other.geom.axes.index_data(axis_name)

        # move axes because numpy dot product is performed on last axis of a and second-to-last axis of b
        data = np.moveaxis(self.data, loc, -1)

        if len(other.geom.axes) > 1:
            other_data = np.moveaxis(other.data[..., 0, 0], other_loc, -2)
        else:
            other_data = other.data[..., 0, 0]

        data = np.dot(data, other_data)

        # prepare new axes with expected shape (i.e. common axis replaced by other's axes)
        index = self.geom.axes.index(axis_name)
        axes1 = self.geom.axes.drop(axis_name)
        inserted_axes = other.geom.axes.drop(axis_name)
        new_axes = axes1[:index] + inserted_axes + axes1[index:]

        # reorder axes to get the expected shape
        remaining_axes = np.arange(len(inserted_axes))
        old_axes_pos = -1 - remaining_axes
        new_axes_pos = loc + remaining_axes[::-1]

        data = np.moveaxis(data, old_axes_pos, new_axes_pos)

        geom = self.geom.to_image().to_cube(new_axes)
        return self._init_copy(geom=geom, data=data)

    def __matmul__(self, other):
        """Apply dot product with the input map.

        The input Map has to share a single MapAxis with the current Map.
        Because it has no spatial dimension, it must be a `~gammapy.maps.RegionNDMap`.
        """
        return self.dot(other)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings

import logging
import numpy as np
from gammapy.datasets import Datasets
from gammapy.datasets.actors import DatasetsActor
from gammapy.modeling import Fit
from gammapy.modeling.selection import NestedModelSelection
from gammapy.modeling.parameter import restore_parameters_status
from gammapy.stats.utils import ts_to_sigma
from gammapy.utils.roots import find_roots
from .core import Estimator

log = logging.getLogger(__name__)

__all__ = ["ParameterEstimator", "ParameterSensitivityEstimator"]


class ParameterEstimator(Estimator):
    """Model parameter estimator.

    Estimates a model parameter for a group of datasets. Compute best fit value,
    symmetric and delta(TS) for a given null value. Additionally asymmetric errors
    as well as parameter upper limit and fit statistic profile can be estimated.

    Parameters
    ----------
    n_sigma : float, optional
        Sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    n_sigma_sensitivity : float, optional
        Sigma to use for sensitivity computation. Must be a positive value.
        Default is 5.
    null_value : float, optional
        Which null value to use for the parameter. Default is 1e-150.
    selection_optional : list of str, optional
        Which additional quantities to estimate. Available options are:

            * "all": all the optional steps are executed.
            * "errn-errp": estimate asymmetric errors on parameter best fit value.
            * "ul": estimate upper limits.
            * "scan": estimate fit statistic profiles.

        Default is None, so the optional steps are not executed.
    fit : `~gammapy.modeling.Fit`, optional
        Fit instance specifying the backend and fit options. If None, the `~gammapy.modeling.Fit` instance is created
        internally. Default is None.
    reoptimize : bool, optional
        Re-optimize other free model parameters. Default is True.

    Examples
    --------
    >>> from gammapy.datasets import SpectrumDatasetOnOff, Datasets
    >>> from gammapy.modeling.models import SkyModel, PowerLawSpectralModel
    >>> from gammapy.estimators import ParameterEstimator
    >>>
    >>> filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    >>> dataset = SpectrumDatasetOnOff.read(filename)
    >>> datasets = Datasets([dataset])
    >>> spectral_model = PowerLawSpectralModel(amplitude="3e-11 cm-2s-1TeV-1", index=2.7)
    >>>
    >>> model = SkyModel(spectral_model=spectral_model, name="Crab")
    >>> model.spectral_model.amplitude.scan_n_values = 10
    >>>
    >>> for dataset in datasets:
    ...     dataset.models = model
    >>>
    >>> estimator = ParameterEstimator(selection_optional="all")
    >>> result = estimator.run(datasets, parameter="amplitude")
    """

    tag = "ParameterEstimator"
    _available_selection_optional = ["errn-errp", "ul", "scan", "sensitivity"]

    def __init__(
        self,
        n_sigma=1,
        n_sigma_ul=2,
        n_sigma_sensitivity=5,
        null_value=1e-150,
        selection_optional=None,
        fit=None,
        reoptimize=True,
    ):
        self.n_sigma = n_sigma
        self.n_sigma_ul = n_sigma_ul
        self.n_sigma_sensitivity = n_sigma_sensitivity
        self.null_value = null_value
        self.selection_optional = selection_optional

        if fit is None:
            fit = Fit()

        self.fit = fit
        self.reoptimize = reoptimize

    def estimate_best_fit(self, datasets, parameter):
        """Estimate parameter asymmetric errors.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        parameter : `~gammapy.modeling.Parameter`
            For which parameter to get the value.

        Returns
        -------
        result : dict
            Dictionary with the various parameter estimation values. Entries are:

                * parameter.name: best fit parameter value.
                * "stat": best fit total stat.
                * "success": boolean flag for fit success.
                * parameter.name_err: covariance-based error estimate on parameter value.
        """
        value, total_stat, success, error = np.nan, 0.0, False, np.nan

        if np.any(datasets.contributes_to_stat):
            result = self.fit.run(datasets=datasets)
            value, error = parameter.value, parameter.error
            total_stat = result.optimize_result.total_stat
            success = result.success

        return {
            f"{parameter.name}": value,
            "stat": total_stat,
            "success": success,
            f"{parameter.name}_err": error * self.n_sigma,
        }

    def estimate_ts(self, datasets, parameter):
        """Estimate parameter ts.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        parameter : `~gammapy.modeling.Parameter`
            For which parameter to get the value.

        Returns
        -------
        result : dict
            Dictionary with the test statistic of the best fit value compared to the null hypothesis. Entries are:

                * "ts" : fit statistic difference with null hypothesis.
                * "npred" : predicted number of counts per dataset.
                * "stat_null" : total stat corresponding to the null hypothesis
        """
        npred = self.estimate_npred(datasets=datasets)

        if not np.any(datasets.contributes_to_stat):
            stat = np.nan
            npred["npred"][...] = np.nan
        else:
            stat = datasets.stat_sum()

        with datasets.parameters.restore_status():
            # compute ts value
            parameter.value = self.null_value

            if self.reoptimize:
                parameter.frozen = True
                _ = self.fit.optimize(datasets=datasets)

            ts = datasets.stat_sum() - stat
            stat_null = datasets.stat_sum()

        return {"ts": ts, "npred": npred["npred"], "stat_null": stat_null}

    def estimate_errn_errp(self, datasets, parameter):
        """Estimate parameter asymmetric errors.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        parameter : `~gammapy.modeling.Parameter`
            For which parameter to get the value.

        Returns
        -------
        result : dict
            Dictionary with the parameter asymmetric errors. Entries are:

                * {parameter.name}_errp : positive error on parameter value.
                * {parameter.name}_errn : negative error on parameter value.
        """
        if not np.any(datasets.contributes_to_stat):
            return {
                f"{parameter.name}_errp": np.nan,
                f"{parameter.name}_errn": np.nan,
            }

        self.fit.optimize(datasets=datasets)

        res = self.fit.confidence(
            datasets=datasets,
            parameter=parameter,
            sigma=self.n_sigma,
            reoptimize=self.reoptimize,
        )

        return {
            f"{parameter.name}_errp": res["errp"],
            f"{parameter.name}_errn": res["errn"],
        }

    def estimate_scan(self, datasets, parameter):
        """Estimate parameter statistic scan.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            The datasets used to estimate the model parameter.
        parameter : `~gammapy.modeling.Parameter`
            For which parameter to get the value.

        Returns
        -------
        result : dict
            Dictionary with the parameter fit scan values. Entries are:

                * parameter.name_scan : parameter values scan.
                * "stat_scan" : fit statistic values scan.
        """
        scan_values = parameter.scan_values

        if not np.any(datasets.contributes_to_stat):
            return {
                f"{parameter.name}_scan": scan_values,
                "stat_scan": scan_values * np.nan,
            }

        self.fit.optimize(datasets=datasets)

        profile = self.fit.stat_profile(
            datasets=datasets, parameter=parameter, reoptimize=self.reoptimize
        )

        return {
            f"{parameter.name}_scan": scan_values,
            "stat_scan": profile["stat_scan"],
        }

    def estimate_ul(self, datasets, parameter):
        """Estimate parameter ul.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            The datasets used to estimate the model parameter.
        parameter : `~gammapy.modeling.Parameter`
            For which parameter to get the value.

        Returns
        -------
        result : dict
            Dictionary with the parameter upper limits. Entries are:

                * parameter.name_ul : upper limit on parameter value.
        """
        if not np.any(datasets.contributes_to_stat):
            return {f"{parameter.name}_ul": np.nan}

        self.fit.optimize(datasets=datasets)

        res = self.fit.confidence(
            datasets=datasets,
            parameter=parameter,
            sigma=self.n_sigma_ul,
            reoptimize=self.reoptimize,
        )
        return {f"{parameter.name}_ul": res["errp"] + parameter.value}

    def estimate_sensitivity(self, datasets, parameter):
        """Estimate norm sensitivity for the flux point.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        result : dict
            Dictionary with an array with one entry per dataset with the sum of the
            masked npred.
        """
        estimator = ParameterSensitivityEstimator(
            parameter, self.null_value, n_sigma=self.n_sigma_sensitivity
        )
        value = estimator.run(datasets)
        return {f"{parameter.name}_sensitivity": value}

    @staticmethod
    def estimate_counts(datasets):
        """Estimate counts for the flux point.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        result : dict
            Dictionary with an array with one entry per dataset with the sum of the
            masked counts.
        """
        counts = []

        for dataset in datasets:
            mask = dataset.mask
            counts.append(dataset.counts.data[mask].sum())

        return {"counts": np.array(counts, dtype=int), "datasets": datasets.names}

    @staticmethod
    def estimate_npred(datasets):
        """Estimate npred for the flux point.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        result : dict
            Dictionary with an array with one entry per dataset with the sum of the
            masked npred.
        """
        npred = []

        for dataset in datasets:
            mask = dataset.mask
            npred.append(dataset.npred().data[mask].sum())

        return {"npred": np.array(npred), "datasets": datasets.names}

    def run(self, datasets, parameter):
        """Run the parameter estimator.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            The datasets used to estimate the model parameter.
        parameter : `str` or `~gammapy.modeling.Parameter`
            For which parameter to run the estimator.

        Returns
        -------
        result : dict
            Dictionary with the various parameter estimation values.
            If used without the optional steps, it contains the following keys:

                * parameter.name : best fit parameter value
                * "stat" : best fit total stat
                * "success" : boolean flag for fit success
                * parameter.name_err: covariance-based error estimate on parameter value
                * "ts" : delta(TS) value
                * "npred" : predicted number of counts per dataset
                * "stat_null" : total stat corresponding to the null hypothesis
                * "counts" : counts value per dataset
                * "datasets" : names of the datasets
        """
        if not isinstance(datasets, DatasetsActor):
            datasets = Datasets(datasets)
        parameter = datasets.parameters[parameter]

        with datasets.parameters.restore_status():
            if not self.reoptimize:
                datasets.parameters.freeze_all()
                parameter.frozen = False

            result = self.estimate_best_fit(datasets, parameter)
            result.update(self.estimate_ts(datasets, parameter))

            if "errn-errp" in self.selection_optional:
                result.update(self.estimate_errn_errp(datasets, parameter))

            if "ul" in self.selection_optional:
                result.update(self.estimate_ul(datasets, parameter))

            if "scan" in self.selection_optional:
                result.update(self.estimate_scan(datasets, parameter))

            if "sensitivity" in self.selection_optional:
                result.update(self.estimate_sensitivity(datasets, parameter))

        result.update(self.estimate_counts(datasets))
        return result


class ParameterSensitivityEstimator:
    """Estimate the sensitivity to a given parameter.

    Computes the TS distribution in the non-null hypothesis using the
    log likelihood of the Asimov dataset (i.e. a dataset with counts = npred)
    and the non-central chi2 distribution.
    Once the TS distribution under the testing hypothesis is known,
    one can compute the required parameter value
    to have 50% of measurements above a given significance threshold.


    Parameters
    ----------
    parameter : `~gammapy.modeling.Parameter`
       Parameter to test
    null_value : float or `~gammapy.modeling.Parameter`
        Value of the parameter for the null hypothesis.
    n_sigma : float, optional
        Number of required significance level. Default is 5.
    n_free_parameters : int, optional
        Number of free parameters. Default is None, which utilises len(parameters).
    rtol : float, optional
        Relative precision of the estimate. Used as a stopping criterion.
        Default is 0.01.
    max_niter : int, optional
        Maximal number of iterations used by the root finding algorithm.
        Default is 100.

    References
    ----------
    * `Cowan et al. (2011), "Asymptotic formulae for likelihood-based tests of new physics"
      <https://arxiv.org/abs/1007.1727>`_

    """

    tag = "ParameterSensitivityEstimator"

    def __init__(
        self,
        parameter,
        null_value,
        n_sigma=5,
        n_free_parameters=None,
        rtol=0.01,
        max_niter=100,
    ):
        self.test = NestedModelSelection(
            [parameter], [null_value], n_free_parameters=n_free_parameters
        )
        self.parameter = parameter
        self.n_sigma = n_sigma
        self.rtol = rtol
        self.max_niter = max_niter

    def _fcn(self, value, datasets):
        """Call the Test Statistics function."""
        self.parameter.value = value
        ts_asimov = self.test.ts_asimov(datasets)
        return ts_to_sigma(ts_asimov, ts_asimov=ts_asimov) - self.n_sigma

    def parameter_matching_significance(self, datasets):
        """Parameter value  matching the target significance."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            roots, res = find_roots(
                self._fcn,
                self.parameter.conf_min,
                self.parameter.conf_max,
                args=(datasets,),
                nbin=100,
                maxiter=self.max_niter,
                rtol=self.rtol,
                points_scale=self.parameter.interp,
            )
        # Where the root finding fails NaN is set as norm
        roots = roots[roots > 0]
        if roots.size > 0:
            return roots[0]
        else:
            return np.nan

    def run(self, datasets):
        """Run the parameter sensitivity estimator.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            The datasets used to estimate the parameter sensitivity.

        Returns
        -------
        result : float
            Parameter sensitivity given as the difference between the value matching
            the target significance and the null value.
        """
        with restore_parameters_status(self.test.parameters):
            value = self.parameter_matching_significance(datasets)

        return value - self.test.null_values[0]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import numpy as np
import astropy.units as u
from astropy.coordinates import Angle
from astropy.coordinates.erfa_astrom import erfa_astrom, ErfaAstromInterpolator
from astropy.table import Table
from astropy.time import Time
from gammapy.data import FixedPointingInfo, PointingMode
from gammapy.irf import EDispMap, FoVAlignment, PSFMap
from gammapy.maps import Map, RegionNDMap, MapAxis
from gammapy.maps.utils import broadcast_axis_values_to_geom
from gammapy.modeling.models import PowerLawSpectralModel
from gammapy.stats import WStatCountsStatistic
from gammapy.utils.coordinates import FoVICRSFrame, FoVAltAzFrame
from gammapy.utils.regions import compound_region_to_regions

__all__ = [
    "make_counts_off_rad_max",
    "make_counts_rad_max",
    "make_edisp_kernel_map",
    "make_edisp_map",
    "make_map_background_irf",
    "make_map_exposure_true_energy",
    "make_psf_map",
    "make_theta_squared_table",
    "make_effective_livetime_map",
    "make_observation_time_map",
]

log = logging.getLogger(__name__)

MINIMUM_TIME_STEP = 1 * u.s  # Minimum time step used to handle FoV rotations
EARTH_ANGULAR_VELOCITY = 360 * u.deg / u.day


def _compute_rotation_time_steps(
    time_start, time_stop, fov_rotation, pointing_altaz, location
):
    """
    Compute the time intervals between start and stop times, such that the FoV associated to a fixed RaDec position rotates
    by 'fov_rotation' in AltAz frame during each time step.
    It assumes that the rotation rate at the provided pointing is a good estimate of the rotation rate over the full
    output duration (first order approximation).


    Parameters
    ----------
    time_start : `~astropy.time.Time`
        Start time
    time_stop : `~astropy.time.Time`
        Stop time
    fov_rotation : `~astropy.units.Quantity`
        Rotation angle.
    pointing_altaz : `~astropy.coordinates.SkyCoord`
        Pointing direction.
    location : `astropy.coordinates.EarthLocation`
        Observatory location
    Returns
    -------
    times : `~astropy.time.Time`
        Times associated with the requested rotation.
    """

    def _time_step(rotation, pnt_altaz):
        denom = (
            EARTH_ANGULAR_VELOCITY
            * np.cos(pnt_altaz.location.lat.rad)
            * np.abs(np.cos(pnt_altaz.az.rad))
        )
        return rotation * np.cos(pnt_altaz.alt.rad) / denom

    time = time_start
    times = [time]
    while time < time_stop:
        time_step = _time_step(fov_rotation, pointing_altaz.get_altaz(time, location))
        time_step = max(time_step, MINIMUM_TIME_STEP)
        time = min(time + time_step, time_stop)
        times.append(time)
    return Time(times)


def make_map_exposure_true_energy(
    pointing, livetime, aeff, geom, use_region_center=True
):
    """Compute exposure map.

    This map has a true energy axis, the exposure is not combined
    with energy dispersion.

    Parameters
    ----------
    pointing : `~astropy.coordinates.SkyCoord`
        Pointing direction.
    livetime : `~astropy.units.Quantity`
        Livetime.
    aeff : `~gammapy.irf.EffectiveAreaTable2D`
        Effective area.
    geom : `~gammapy.maps.WcsGeom`
        Map geometry (must have an energy axis).
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.

    Returns
    -------
    map : `~gammapy.maps.WcsNDMap`
        Exposure map.
    """
    if isinstance(pointing, FixedPointingInfo):
        origin = pointing.get_icrs(pointing.obstime)
    else:
        origin = pointing

    fov_frame = FoVICRSFrame(origin=origin)

    exposure = project_irf_on_geom(geom, aeff, fov_frame, use_region_center)

    exposure *= u.Quantity(livetime)
    exposure = exposure.to_unit("m2 s")
    exposure.meta.update({"livetime": livetime, "is_pointlike": aeff.is_pointlike})

    return exposure


def _map_spectrum_weight(map, spectrum=None):
    """Weight a map with a spectrum.

    This requires map to have an "energy" axis.
    The weights are normalised so that they sum to 1.
    The mean and unit of the output image is the same as of the input cube.

    At the moment this is used to get a weighted exposure image.

    Parameters
    ----------
    map : `~gammapy.maps.Map`
        Input map with an "energy" axis.
    spectrum : `~gammapy.modeling.models.SpectralModel`, optional
        Spectral model to compute the weights.
        Default is None, which is a power-law with spectral index of 2.

    Returns
    -------
    map_weighted : `~gammapy.maps.Map`
        Weighted image.
    """
    if spectrum is None:
        spectrum = PowerLawSpectralModel(index=2.0)

    # Compute weights vector
    for name in map.geom.axes.names:
        if "energy" in name:
            energy_name = name
    energy_edges = map.geom.axes[energy_name].edges
    weights = spectrum.integral(
        energy_min=energy_edges[:-1], energy_max=energy_edges[1:]
    )
    weights /= weights.sum()
    shape = np.ones(len(map.geom.data_shape))
    shape[0] = -1
    return map * weights.reshape(shape.astype(int))


def make_map_background_irf(
    pointing,
    ontime,
    bkg,
    geom,
    time_start,
    fov_rotation_step=1.0 * u.deg,
    oversampling=None,
    use_region_center=True,
    location=None,
):
    """Compute background map from background IRFs.

    Parameters
    ----------
    pointing : `~gammapy.data.FixedPointingInfo` or `~astropy.coordinates.SkyCoord`
        Observation pointing.

        - If a `~gammapy.data.FixedPointingInfo` is passed, FOV coordinates
          are properly computed.
        - If a `~astropy.coordinates.SkyCoord` is passed, FOV frame rotation
          is not taken into account.

    ontime : `~astropy.units.Quantity`
        Observation ontime. i.e. not corrected for deadtime
        see https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/bkg/index.html#notes)  # noqa: E501
    bkg : `~gammapy.irf.Background3D`
        Background rate model.
    geom : `~gammapy.maps.WcsGeom`
        Reference geometry.
    time_start : `~astropy.time.Time`
        Observation start time.
    fov_rotation_step : `~astropy.units.Quantity`
        Maximum rotation error to create sub-time bins if the irf is 3D and in AltAz.
        Default is 1.0 deg.
    oversampling : int, optional
        Oversampling factor in energy, used for the background model evaluation.
        Default is None.
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.
    location : `astropy.coordinates.EarthLocation`, optional
        Observatory location
    Returns
    -------
    background : `~gammapy.maps.WcsNDMap`
        Background predicted counts sky cube in reconstructed energy.
    """
    # TODO:
    #  This implementation can be improved in one way:
    #  Use the pointing table (does not currently exist in CTA files) to
    #  obtain the RA DEC and time for each interval. This then considers that
    #  the pointing might change slightly over the observation duration
    # Compute intermediate time ranges if needed
    times = Time([time_start, time_start + ontime])

    if not bkg.has_offset_axis and bkg.fov_alignment == FoVAlignment.ALTAZ:
        if not isinstance(pointing, FixedPointingInfo):
            raise TypeError(
                "make_map_background_irf requires FixedPointingInfo if "
                "BackgroundIRF.fov_alignement is ALTAZ",
            )
        times = _compute_rotation_time_steps(
            time_start, time_start + ontime, fov_rotation_step, pointing, location
        )
        origin = pointing.get_altaz(times, location)
        fov_frame = FoVAltAzFrame(
            origin=origin, location=origin.location, obstime=times
        )
    else:
        if isinstance(pointing, FixedPointingInfo):
            if pointing.mode == PointingMode.POINTING:
                origin = pointing.fixed_icrs
            else:
                raise NotImplementedError(
                    "Drift pointing mode is not supported for background calculation."
                )
        else:
            origin = pointing

        fov_frame = FoVICRSFrame(origin=origin)

    if oversampling is not None:
        geom = geom.upsample(factor=oversampling, axis_name="energy")

    bkg_map = integrate_project_irf_on_geom(geom, bkg, fov_frame, use_region_center)
    bkg_map *= ontime

    if oversampling is not None:
        bkg_map = bkg_map.downsample(factor=oversampling, axis_name="energy")

    return bkg_map.to_unit("")


def make_psf_map(psf, pointing, geom, exposure_map=None):
    """Make a PSF map for a single observation.

    Expected axes : rad and true energy in this specific order.
    The name of the rad MapAxis is expected to be 'rad'.

    Parameters
    ----------
    psf : `~gammapy.irf.PSF3D`
        The PSF IRF.
    pointing : `~astropy.coordinates.SkyCoord`
        The pointing direction.
    geom : `~gammapy.maps.Geom`
        The map geometry to be used. It provides the target geometry.
        rad and true energy axes should be given in this specific order.
    exposure_map : `~gammapy.maps.Map`, optional
        The associated exposure map.
        Default is None.

    Returns
    -------
    psfmap : `~gammapy.irf.PSFMap`
        The resulting PSF map.
    """
    if isinstance(pointing, FixedPointingInfo):
        origin = pointing.get_icrs(pointing.obstime)
    else:
        origin = pointing

    fov_frame = FoVICRSFrame(origin=origin)

    psf_map = project_irf_on_geom(geom, psf, fov_frame)
    psf_map.normalize(axis_name="rad")
    return PSFMap(psf_map, exposure_map)


def make_edisp_map(edisp, pointing, geom, exposure_map=None, use_region_center=True):
    """Make an edisp map for a single observation.

    Expected axes : migra and true energy in this specific order.
    The name of the migra MapAxis is expected to be 'migra'.

    Parameters
    ----------
    edisp : `~gammapy.irf.EnergyDispersion2D`
        The 2D energy dispersion IRF.
    pointing : `~astropy.coordinates.SkyCoord`
        The pointing direction.
    geom : `~gammapy.maps.Geom`
        The map geometry to be used. It provides the target geometry.
        migra and true energy axes should be given in this specific order.
    exposure_map : `~gammapy.maps.Map`, optional
        The associated exposure map.
        Default is None.
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.

    Returns
    -------
    edispmap : `~gammapy.irf.EDispMap`
        The resulting energy dispersion map.
    """
    if isinstance(pointing, FixedPointingInfo):
        origin = pointing.get_icrs(pointing.obstime)
    else:
        origin = pointing

    fov_frame = FoVICRSFrame(origin=origin)

    edisp_map = project_irf_on_geom(geom, edisp, fov_frame).to_unit("")
    edisp_map.normalize(axis_name="migra")
    return EDispMap(edisp_map, exposure_map)


def make_edisp_kernel_map(
    edisp, pointing, geom, exposure_map=None, use_region_center=True
):
    """Make an edisp kernel map for a single observation.

    Expected axes : (reco) energy and true energy in this specific order.
    The name of the reco energy MapAxis is expected to be 'energy'.
    The name of the true energy MapAxis is expected to be 'energy_true'.

    Parameters
    ----------
    edisp : `~gammapy.irf.EnergyDispersion2D`
        The 2D energy dispersion IRF.
    pointing : `~astropy.coordinates.SkyCoord`
        The pointing direction.
    geom : `~gammapy.maps.Geom`
        The map geometry to be used. It provides the target geometry.
        energy and true energy axes should be given in this specific order.
    exposure_map : `~gammapy.maps.Map`, optional
        The associated exposure map.
        Default is None.
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.

    Returns
    -------
    edispmap : `~gammapy.irf.EDispKernelMap`
        the resulting EDispKernel map
    """
    # Use EnergyDispersion2D migra axis.
    migra_axis = edisp.axes["migra"]

    # Create temporary EDispMap Geom
    new_geom = geom.to_image().to_cube([migra_axis, geom.axes["energy_true"]])

    edisp_map = make_edisp_map(
        edisp, pointing, new_geom, exposure_map, use_region_center
    )

    return edisp_map.to_edisp_kernel_map(geom.axes["energy"])


def make_theta_squared_table(
    observations, theta_squared_axis, position, position_off=None, energy_edges=None
):
    """Make theta squared distribution in the same FoV for a list of `~gammapy.data.Observation` objects.

    The ON theta2 profile is computed from a given distribution, on_position.
    By default, the OFF theta2 profile is extracted from a mirror position
    radially symmetric in the FOV to pos_on.

    The ON and OFF regions are assumed to be of the same size, so the normalisation
    factor between both region alpha = 1.

    Parameters
    ----------
    observations: `~gammapy.data.Observations`
        List of observations.
    theta_squared_axis : `~gammapy.maps.MapAxis`
        Axis of edges of the theta2 bin used to compute the distribution.
    position : `~astropy.coordinates.SkyCoord`
        Position from which the on theta^2 distribution is computed.
    position_off : `astropy.coordinates.SkyCoord`
        Position from which the OFF theta^2 distribution is computed.
        Default is reflected position w.r.t. to the pointing position.
    energy_edges : list of `~astropy.units.Quantity`, optional
        Edges of the energy bin where the theta squared distribution
        is evaluated. For now, only one interval is accepted.
        Default is None.

    Returns
    -------
    table : `~astropy.table.Table`
        Table containing the on counts, the off counts, acceptance, off acceptance and alpha
        for each theta squared bin.
    """
    if not theta_squared_axis.edges.unit.is_equivalent("deg2"):
        raise ValueError("The theta2 axis should be equivalent to deg2")

    table = Table()

    table["theta2_min"] = theta_squared_axis.edges[:-1]
    table["theta2_max"] = theta_squared_axis.edges[1:]
    table["counts"] = 0
    table["counts_off"] = 0
    table["acceptance"] = 0.0
    table["acceptance_off"] = 0.0

    alpha_tot = np.zeros(len(table))
    livetime_tot = 0

    create_off = position_off is None

    if energy_edges is not None:
        if len(energy_edges) == 2:
            table.meta["Energy_filter"] = energy_edges
        else:
            raise ValueError(
                f"Only  supports one energy interval but {len(energy_edges) - 1} passed."
            )

    for observation in observations:
        events = observation.events
        if energy_edges is not None:
            events = events.select_energy(energy_range=energy_edges)

        event_position = events.radec
        pointing = observation.get_pointing_icrs(observation.tmid)

        separation = position.separation(event_position)
        counts, _ = np.histogram(separation**2, theta_squared_axis.edges)
        table["counts"] += counts

        if create_off:
            # Estimate the position of the mirror position
            pos_angle = pointing.position_angle(position)
            sep_angle = pointing.separation(position)
            position_off = pointing.directional_offset_by(
                pos_angle + Angle(np.pi, "rad"), sep_angle
            )

        # Angular distance of the events from the mirror position
        separation_off = position_off.separation(event_position)

        # Extract the ON and OFF theta2 distribution from the two positions.
        counts_off, _ = np.histogram(separation_off**2, theta_squared_axis.edges)
        table["counts_off"] += counts_off

        # Normalisation between ON and OFF is one
        acceptance = np.ones(theta_squared_axis.nbin)
        acceptance_off = np.ones(theta_squared_axis.nbin)

        table["acceptance"] += acceptance
        table["acceptance_off"] += acceptance_off
        alpha = acceptance / acceptance_off
        alpha_tot += alpha * observation.observation_live_time_duration.to_value("s")
        livetime_tot += observation.observation_live_time_duration.to_value("s")

    alpha_tot /= livetime_tot
    table["alpha"] = alpha_tot

    stat = WStatCountsStatistic(table["counts"], table["counts_off"], table["alpha"])
    table["excess"] = stat.n_sig
    table["sqrt_ts"] = stat.sqrt_ts
    table["excess_errn"] = stat.compute_errn()
    table["excess_errp"] = stat.compute_errp()

    table.meta["ON_RA"] = position.icrs.ra
    table.meta["ON_DEC"] = position.icrs.dec
    return table


def make_counts_rad_max(geom, rad_max, events):
    """Extract the counts using for the ON region size the values in the `RAD_MAX_2D` table.

    Parameters
    ----------
    geom : `~gammapy.maps.RegionGeom`
        Reference map geometry.
    rad_max : `~gammapy.irf.RadMax2D`
        Rhe RAD_MAX_2D table IRF.
    events : `~gammapy.data.EventList`
        Event list to be used to compute the ON counts.

    Returns
    -------
    counts : `~gammapy.maps.RegionNDMap`
        Counts vs estimated energy extracted from the ON region.
    """
    selected_events = events.select_rad_max(
        rad_max=rad_max, position=geom.region.center
    )

    counts = Map.from_geom(geom=geom)
    counts.fill_events(selected_events)
    return counts


def make_counts_off_rad_max(geom_off, rad_max, events):
    """Extract the OFF counts from a list of point regions and given rad max.

    This method does **not** check for overlap of the regions defined by rad_max.

    Parameters
    ----------
    geom_off : `~gammapy.maps.RegionGeom`
        Reference map geometry for the on region.
    rad_max : `~gammapy.irf.RadMax2D`
        The RAD_MAX_2D table IRF.
    events : `~gammapy.data.EventList`
        Event list to be used to compute the OFF counts.

    Returns
    -------
    counts_off : `~gammapy.maps.RegionNDMap`
        OFF Counts vs estimated energy extracted from the ON region.
    """
    if not geom_off.is_all_point_sky_regions:
        raise ValueError(
            f"Only supports PointSkyRegions, got {geom_off.region} instead"
        )

    counts_off = RegionNDMap.from_geom(geom=geom_off)

    for off_region in compound_region_to_regions(geom_off.region):
        selected_events = events.select_rad_max(
            rad_max=rad_max, position=off_region.center
        )
        counts_off.fill_events(selected_events)

    return counts_off


def make_observation_time_map(observations, geom, offset_max=None):
    """Compute the total observation time on the target geometry for a list of observations.

    Parameters
    ----------
    observations : `~gammapy.data.Observations`
            Observations container containing list of observations.
    geom : `~gammapy.maps.Geom`
            Reference geometry.
    offset_max : `~astropy.units.quantity.Quantity`, optional
        The maximum offset FoV. Default is None.
        If None, it will be taken from the IRFs.

    Returns
    -------
    exposure : `~gammapy.maps.Map`
        Effective livetime.
    """
    geom = geom.to_image()
    stacked = Map.from_geom(geom, unit=u.h)
    for obs in observations:
        if offset_max is None:
            offset_max = guess_instrument_fov(obs)
        coords = geom.get_coord(sparse=True)
        offset = coords.skycoord.separation(obs.get_pointing_icrs(obs.tmid))
        mask = offset < offset_max
        c1 = coords.apply_mask(mask)
        weights = np.ones(c1.shape) * obs.observation_live_time_duration
        stacked.fill_by_coord(coords=c1, weights=weights)
    return stacked


def make_effective_livetime_map(observations, geom, offset_max=None):
    """Compute the acceptance corrected livetime map for a list of observations.

    Parameters
    ----------
    observations : `~gammapy.data.Observations`
        Observations container containing list of observations.
    geom : `~gammapy.maps.Geom`
        Reference geometry.
    offset_max : `~astropy.units.quantity.Quantity`, optional
        The maximum offset FoV. Default is None.

    Returns
    -------
     exposure : `~gammapy.maps.Map`
        Effective livetime.
    """
    livetime = Map.from_geom(geom, unit=u.hr)
    for obs in observations:
        if offset_max is None:
            offset_max = guess_instrument_fov(obs)
        geom_obs = geom.cutout(
            position=obs.get_pointing_icrs(obs.tmid), width=2.0 * offset_max
        )
        coords = geom_obs.get_coord()
        offset = coords.skycoord.separation(obs.get_pointing_icrs(obs.tmid))
        mask = offset < offset_max

        exposure = make_map_exposure_true_energy(
            pointing=geom.center_skydir,
            livetime=obs.observation_live_time_duration,
            aeff=obs.aeff,
            geom=geom_obs,
            use_region_center=True,
        )

        on_axis = obs.aeff.evaluate(
            offset=0.0 * u.deg, energy_true=geom.axes["energy_true"].center
        )
        on_axis = on_axis.reshape((on_axis.shape[0], 1, 1))
        lv_obs = exposure * mask / on_axis
        livetime.stack(lv_obs)
    return livetime


def guess_instrument_fov(obs):
    """
    Guess the camera field of view for the given observation
    from the IRFs. This simply takes the maximum offset of the
    effective area IRF.
    TODO: This logic will break for more complex IRF models.
    A better option would be to compute the offset at which
    the effective area is above 10% of the maximum.

    Parameters
    ----------
    obs : `~gammapy.data.Observation`
        Observation container.

    Returns
    -------
    offset_max : `~astropy.units.quantity.Quantity`
        The maximum offset of the effective area IRF.
    """
    if "aeff" not in obs.available_irfs:
        raise ValueError("No Effective area IRF to infer the FoV from")
    if obs.aeff.is_pointlike:
        raise ValueError("Cannot guess FoV from pointlike IRFs")
    if "offset" not in obs.aeff.axes.names:
        raise ValueError("Offset axis not present!")
    return obs.aeff.axes["offset"].center[-1]


def _get_fov_coord(
    skycoord, fov_frame, use_offset=True, reverse_lon=False, time_resolution=1000 * u.s
):
    """Return coord dict in fov_coord."""
    coords = {}

    if use_offset:
        coords["offset"] = skycoord.separation(fov_frame.origin)
    else:
        sign = -1.0 if reverse_lon else 1.0

        with erfa_astrom.set(ErfaAstromInterpolator(time_resolution)):
            fov_coords = skycoord.transform_to(fov_frame)

        if len(fov_frame.shape) == 1:
            coords["fov_lon"] = np.moveaxis(fov_coords.fov_lon, -1, 0)
            coords["fov_lat"] = np.moveaxis(fov_coords.fov_lat, -1, 0)
        else:
            coords["fov_lon"] = sign * fov_coords.fov_lon
            coords["fov_lat"] = fov_coords.fov_lat

    return coords


def project_irf_on_geom(geom, irf, fov_frame, use_region_center=True):
    """Evaluate and project an IRF on a given `~gammapy.maps.Geom` object according to a given FoV Frame.

    When ``geom`` is a `~gammapy.maps.RegionGeom`, the IRF is evaluated at the region center when
    ``user_region_center is True``. Otherwise, the IRF is evaluated and averaged over the whole region.

    Parameters
    ----------
    geom : `~gammapy.maps.Geom`
        Geometry to project on. It must follow the required axes of the input IRF.
    irf : `~gammapy.itf.IRF`
        IRF to reproject.
    fov_frame : `~gammapy.utils.coordinate.FoVICRSFrame` or `~gammapy.utils.coordinate.FoVAltAzFrame`
        FoV frame to convert geometry to FoV coordinates.
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.

    Returns
    -------
    map : `~gammapy.maps.Map`
        Map containing the projected IRF.
    """
    if not use_region_center:
        image_geom = geom.to_wcs_geom().to_image()
        region_coord, weights = geom.get_wcs_coord_and_weights()
        skycoord = region_coord.skycoord
    else:
        image_geom = geom.to_image()
        skycoord = image_geom.get_coord().skycoord

    coords = _get_fov_coord(skycoord, fov_frame, irf.has_offset_axis)

    non_spatial_axes = set(irf.required_arguments) - set(
        ["offset", "fov_lon", "fov_lat"]
    )

    for axis_name in non_spatial_axes:
        coords[axis_name] = broadcast_axis_values_to_geom(geom, axis_name)

    data = irf.evaluate(**coords)
    if not use_region_center:
        data = np.average(data, axis=-1, weights=weights, keepdims=True)

    return Map.from_geom(geom=geom, data=data.value, unit=data.unit)


def integrate_project_irf_on_geom(geom, irf, fov_frame, use_region_center=True):
    """Integrate and project an IRF on a given `~gammapy.maps.Geom` object according to a given FoV Frame.

    The IRF is integrated in energy and multiplied by the solid angle.

    When ``geom`` is a `~gammapy.maps.RegionGeom`, the IRF is evaluated at the region center when
    ``user_region_center is True``. Otherwise, the IRF is evaluated and averaged over the whole region.

    Parameters
    ----------
    geom : `~gammapy.maps.Geom`
        Geometry to project on. It must follow the required axes of the input IRF.
    irf : `~gammapy.irf.BackgroundIRF`
        IRF to reproject. Typically a background IRF.
    fov_frame : `~gammapy.utils.coordinate.FoVICRSFrame` or `~gammapy.utils.coordinate.FoVAltAzFrame`
        FoV frame to convert geometry to FoV coordinates.
    use_region_center : bool, optional
        For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
        If False, average over the whole region.
        Default is True.

    Returns
    -------
    map : `~gammapy.maps.Map`
        Map containing the projected IRF.
    """
    from scipy.integrate import trapezoid

    if not use_region_center:
        image_geom = geom.to_wcs_geom().to_image()
        region_coord, weights = geom.get_wcs_coord_and_weights()
        skycoord = region_coord.skycoord
    else:
        image_geom = geom.to_image()
        skycoord = image_geom.get_coord().skycoord

    new_geom = geom
    # In case we need to integrate over time
    if len(fov_frame.shape) == 1:
        skycoord = skycoord[..., np.newaxis]

        # Assume ordered times
        time = MapAxis.from_edges(
            (fov_frame.obstime - fov_frame.obstime[0]).to_value("s"),
            unit="s",
            name="time",
        )
        axes = geom.axes

        new_geom = image_geom.to_cube([time, *axes])

    reverse_lon = irf.fov_alignment == "REVERSE_LON_RADEC"
    coords = _get_fov_coord(skycoord, fov_frame, irf.has_offset_axis, reverse_lon)

    non_spatial_axes = set(irf.required_arguments) - set(
        ["offset", "fov_lon", "fov_lat"]
    )

    for axis_name in non_spatial_axes:
        coords[axis_name] = broadcast_axis_values_to_geom(new_geom, axis_name, False)

    data = irf.integrate_log_log(**coords, axis_name="energy")

    if len(fov_frame.shape) == 1:
        time = new_geom.axes["time"]
        ontime = time.bin_width.sum().to_value("s")
        delta = np.reshape(time.edges.to_value("s"), (1, time.nbin + 1, 1, 1))
        data = trapezoid(data, delta, axis=1) / ontime

    if use_region_center:
        data *= image_geom.solid_angle()
    else:
        idx = image_geom.coord_to_idx(region_coord)
        data *= image_geom.solid_angle().T[idx]
        data = np.sum(weights * data, axis=2, keepdims=True)

    return Map.from_geom(geom=geom, data=data.value, unit=data.unit)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from .likelihood import Likelihood

__all__ = ["optimize_sherpa", "covariance_sherpa"]


def get_sherpa_optimizer(name):
    from sherpa.optmethods import GridSearch, LevMar, MonCar, NelderMead

    return {
        "levmar": LevMar,
        "simplex": NelderMead,
        "moncar": MonCar,
        "gridsearch": GridSearch,
    }[name]()


class SherpaLikelihood(Likelihood):
    """Likelihood function interface for Sherpa."""

    def fcn(self, factors):
        self.parameters.set_parameter_factors(factors)
        total_stat = self.function()

        if self.store_trace:
            self.store_trace_iteration(total_stat)

        return total_stat, 0


def optimize_sherpa(parameters, function, store_trace=False, **kwargs):
    """Sherpa optimization wrapper method.

    Parameters
    ----------
    parameters : `~gammapy.modeling.Parameters`
        Parameter list with starting values.
    function : callable
        Likelihood function.
    **kwargs : dict
        Options passed to the optimizer instance.

    Returns
    -------
    result : (factors, info, optimizer)
        Tuple containing the best fit factors, some information and the optimizer instance.
    """
    method = kwargs.pop("method", "simplex")
    optimizer = get_sherpa_optimizer(method)
    optimizer.config.update(kwargs)

    pars = [par.factor for par in parameters.free_parameters]
    parmins = np.nan_to_num(
        [par.factor_min for par in parameters.free_parameters], nan=-np.inf
    )
    parmaxes = np.nan_to_num(
        [par.factor_max for par in parameters.free_parameters], nan=np.inf
    )

    statfunc = SherpaLikelihood(function, parameters, store_trace)

    with np.errstate(invalid="ignore"):
        result = optimizer.fit(
            statfunc=statfunc.fcn, pars=pars, parmins=parmins, parmaxes=parmaxes
        )

    factors = result[1]
    info = {
        "success": result[0],
        "message": result[3],
        "nfev": result[4]["nfev"],
        "trace": statfunc.trace,
    }

    return factors, info, optimizer


def covariance_sherpa():
    raise NotImplementedError
